from copy import deepcopy
from math import pi
from typing import List, Any
import math

# Owned
from hisim.component import Component, SingleTimeStepValues, ComponentInput, ComponentOutput
from hisim import loadtypes as lt
from hisim.simulationparameters import SimulationParameters
from hisim.components.configuration import PhysicsConfig
from hisim.components.configuration import WarmWaterStorageConfig
import hisim.log as log
#from components.extended_storage import WaterSlice

class WaterSlice:

    def check_units(self):
        """
        Check for the correct input values. Wrong value types can lead to a wrong calculation.
        Values to check:
            - all: must be float or int values
            - all: must values be positive
            - height & temperature: can be 0 --> creation of an empty slice if system is not operating
            - temperature cant be 0 if height is bigger than 0. --> frozen slice
            - temperature: must be below 100°C --> boiling point of water
        :return: raise error if needed
        """
        for value in (self.diameter, self.height, self.temperature):
            if not isinstance(value, (int, float)):
                raise TypeError
            # if isinstance(value, bool):   # Otherwise bool results into 0 or 1
            #   raise TypeError
        if self.diameter <= 0 or self.temperature < 0 or self.height < 0:  # or self.temperature == 0 and self.height > 0:
            log.error("Incorrect tank diameter: " + str(self.diameter))
            log.error("Incorrect water temperature: " + str(self.temperature))
            log.error("Incorrect slice height: " + str(self.height))
            raise ValueError
        if self.temperature >= 100:
            log.error("Incorrect water temperature: " + str(self.temperature))
            log.error("Temperature should be below 100°C -> boiling point of water")
            raise ValueError

        if self.enthalpy < 0:
            log.error("Incorrect water enthalpy: " + str(self.enthalpy))
            log.error("The Enthalpy can't be negative")
            raise ValueError

    def __init__(self, tank_diameter: float, height: float, temperature: float):
        """
        ToDo: Give area as input value provided by the storage. This will decrease the calculation time.
        Initialize a new water slice.
        So far, the tank diameter and the slice diameter are the same. Could be changed later on.
        The diameter is consistent for all slices in the tank!
        The height input is also the definition for the slices mass.
        The area is defined by the diameter.
        :param tank_diameter:
        :param height:
        :param temperature:
        """
        self.diameter = tank_diameter
        self.area = math.pi/4 * tank_diameter ** 2
        self.height = height
        self.temperature = temperature
        self.density = PhysicsConfig.water_density
        self.specific_heat_capacity = PhysicsConfig.water_specific_heat_capacity_in_joule_per_kilogram_per_kelvin
        self.mass = self.height * self.area * self.density  # [kg]
        # Enthalpy referencing to the temperature of 0 °C.
        # J = Ws = kg * J/kg*K * K
        self.enthalpy = self.mass * self.specific_heat_capacity * self.temperature
        self.check_units()

    @staticmethod
    def init_from_another_slice( other_slice):
        ws = WaterSlice(other_slice.diameter, other_slice.height, other_slice.temperature )
        return ws
    """
    This is not used
    def calculate_mass(self):
        self.mass = self.height * self.area * self.density
        return self.mass

    def calculate_enthalpy(self):
        self.enthalpy = self.mass * self.specific_heat_capacity * self.temperature
        return self.enthalpy

    def calculate_temperature(self):
        self.temperature = self.enthalpy / (self.mass * self.specific_heat_capacity)
        return self.temperature
    """

    def add_another_slice(self, other_slice):  # this should be -> other_slice: WaterSlice -> but got an error
        """
        If slices are mixed they get a common mass, height, temperature & enthalpy
        The temperature must be changed first because the original masses are necessary
        !! Don't change order of calculations. Temperature must be done first
        -> deleting the other slice must be done manually, because its not always mandatory
        :param other_slice: the slice which should be mixed with the 'self' slice
        :return: no return (changes in self.__) --> Parameters of newly mixed slice are set
        """
        self.temperature = (self.mass * self.temperature + other_slice.mass * other_slice.temperature) / (self.mass + other_slice.mass)
        self.height += other_slice.height
        self.mass += other_slice.mass
        # The new enthalpy is based in the new parameters
        # ToDo: Idea: self.enthalpy = self.enthalpy + other_slice.enthalpy -> should work
        self.enthalpy = self.mass * self.specific_heat_capacity * self.temperature

    def heat_losses_horizontal(self, u_value_tank: Any, seconds_per_timestep: float = 1, ambient_temperature: float = 20)-> Any:
        """
        All slices in the tank must be considered
        ->  this function is callable by the WarmWaterStorage class to apply it on all the slices in the tank
        There are no temperature differences inside one slice, due to the losses taking part on the outside
        Temperature and enthalpy of the corresponding slices will be changed
        Vertical losses are considered in 'def heat_losses_vertical_top_bottom'
        :param u_value_tank:        u-value of the tank [W/m^2K]
        :param seconds_per_timestep:            seconds per timestep [s]
        :param ambient_temperature: temperature around the storage [°C]
        :return: Horizontal energy losses in this timestep [W]
        """
        perimeter_tank = math.pi * self.diameter
        loss_area = perimeter_tank * self.height
        # W/m^2*K * m^2 * K * s = Ws
        energy_losses = u_value_tank * loss_area * (self.temperature - ambient_temperature) * seconds_per_timestep
        self.enthalpy -= energy_losses
        self.temperature = self.enthalpy / (self.mass * self.specific_heat_capacity)
        return energy_losses    # one slice is loosing this amount

    def heat_losses_vertical_top_or_bottom(self, u_value_tank: float, seconds_per_timestep: float = 1, ambient_temperature: float = 20) -> float:
        """
        Heat losses through top OR bottom of the tank. Function has to be called twice (one each)
        Temperature and enthalpy of the corresponding slice will be changed
        :param u_value_tank:        u-value of the tank [W/m^2K]
        :param seconds_per_timestep:            seconds per timestep [s]
        :param ambient_temperature: temperature around the storage [°C]
        :return: Vertical energy losses through top or bottom in this timestep [W]
        """
        energy_losses = u_value_tank * self.area * (self.temperature - ambient_temperature) * seconds_per_timestep
        if energy_losses > self.enthalpy:
            # ToDO: Merge slices
            pass
        self.enthalpy -= energy_losses
        self.temperature = self.enthalpy / (self.mass * self.specific_heat_capacity)
        return energy_losses

    def change_slice_parameters(self, new_temperature: float = -12345.67, new_enthalpy: float = -12345.67, new_mass: float = -12345.67) -> None:
        """
        ToDo: Wie kann man nur eine einzelne Zeile ausführen --> Zweck: wenn anderen werte nicht eingegeben werden / geändert werden sollen
        # Gibt es einen Besseren Ansatz als diesen?
        ToDo: Wenn Temperaturändeurng dann auch Enthalpieänderung; wenn Massenändeung dann auch Enthalpieänderung ; wenn Enthalpieänderung dann Massden UND/ODER Temperaturänderung
        Change the parameters a water slice has. This can be used to heat up / cool down a specific slice in the CHP or the load
        e.g.:Heat up a water slice with a given energy --> Changes in temperature and enthalpy needed
        """
        if new_temperature != -12345.67:
            self.temperature = new_temperature
        if new_enthalpy != -12345.67:
            self.enthalpy = new_enthalpy
        if new_mass != -12345.67:
            self.mass = new_mass

    def calculate_new_temperature(self):
        self.temperature = self.enthalpy / (self.mass * self.specific_heat_capacity)

class WarmWaterStorageSimulation:
    """
    The most important function is simulate_one_timestep. It calls all other functions needed.

    The storage is defined by its diameter, height and temperature. Inside there are water slices represented by my_slices
    """
    def __init__(self, config: WarmWaterStorageConfig) -> None:
        # Initialises a starting tank and adds one slice with the tanks height and the starting temperature
        self.diameter = config.tank_diameter
        self.height_storage = config.tank_height
        self.start_temperature = config.tank_start_temperature

        self.check_units_wws()

        self.area = (pi / 4) * (self.diameter ** 2)
        self.volume = self.area * self.height_storage     # [m^3]
        self.my_slices = []
        self.my_slices.append(WaterSlice(self.diameter, self.height_storage, self.start_temperature))

    def check_units_wws(self):
        """
        Checking the storage values for a correct setup of the tank
        """
        for value in (self.diameter, self.height_storage, self.start_temperature):
            if not isinstance(value, (int, float)):
                raise TypeError
            # if isinstance(value, bool):
                # Otherwise bool results into 0 or 1
            #   raise TypeError
            if value <= 0:
                # The tank must always have positive values. Otherwise it wouldn't be a tank
                raise ValueError
        if self.start_temperature >= 100:
            raise ValueError  # -> Boiling Water

    def begin_new_timestep_alternative(self) -> None:
        """ try to write an alternative to deepcopy"""
        slices_to_save = len(self.my_slices)
        save_slices_step = []
        for i in range(slices_to_save):
            save_slices_step.append(WaterSlice.init_from_another_slice(self.my_slices[i]))
            # Todo: auf restore anwenden

    def reset_to_last_timestep_alternative(self, save_values_step_1:Any) -> None:
        """ try to write an alternative to deepcopy"""
        amount_of_saved_slices = len(save_values_step_1)
        for i in range(amount_of_saved_slices):
            self.my_slices[i] = save_values_step_1[i]

    def begin_new_timestep(self) -> Any:
        """
        Deep copy of my_slices
        -> relevant for framework. This allows to reset to the previous state
        """
        save_values_step_1 = deepcopy(self.my_slices)
        return save_values_step_1

    def reset_to_last_timestep(self, save_values_step_1: Any) -> Any:
        """
        Get back the step before
        Use together with def begin_new_timestep
        """
        self.my_slices = deepcopy(save_values_step_1)
        return

    def create_water_slice(self, slice_temperature_input: float, slice_mass_input: float) -> Any:
        """
        Creation of one water slice.
        The height has to be calculated fom the input mass.
        :param slice_temperature_input [°C]     Temperature for the new water slice
        :param slice_mass_input [kg]            Mass for the new water slice
        :return ws                              New water slice
        """

        height = slice_mass_input / (self.area * 1000)
        temperature = slice_temperature_input
        ws = WaterSlice(self.diameter, height, temperature)
        return ws

    def insert_slice(self, ws: Any, is_from_top: Any) -> Any:
        """
        Insert the created water slice into the tank. The variable is_from_top decides about where to input the slice.
        If True then its added on top, otherwise ist added at the bottom.
        True -> From CHP / probably hot water / input on top
        False -> From Load / probably cold water is coming back // input at the bottom
        What happens with the slice?
        a) is_from_top = True
        b) is_from_top = False
        Option 1:
        a) New slice is added above the original top slice if T_new >= T_0 + 0,5°C
        b) New slice is added below the original bottom slice if T_new <= T_0 - 0,5°C
        Option 2:
        a) New slice is mixed with the top slice if T_new < T_0 + 0,5°C
        b) New slice is mixed with the bottom slice if T_new > T_0 - 0,5°C

        Checking the temperature gradient (--> T_top < T_bottom)
        Slice will be mixed with the next one until the temperature gradient is correct
        - Its enough to only check the first and second slide. The temperature gradient of the other layers should be intact due to the previous steps.
        - this is not relevant if there is only one slice -> len() > 1
        - after mixing slice a) [1] or b) [-1] will be removed

        :param ws:          water slice which should be inserted
        :param is_from_top: where the new slice should be added
        :return:            not return, changes are made with self.___
        """
        temperature_difference = WarmWaterStorageConfig.temperature_difference
        if ws.height == 0:
            # no empty slices will be inserted into the tank
            pass
        else:
            if is_from_top:
                outer_slice = self.my_slices[0]
                if ws.temperature >= (outer_slice.temperature + temperature_difference):
                    self.my_slices.insert(0, ws)
                else:
                    outer_slice.add_another_slice(ws)
                    # Check temperature profile
                    # Only the first two slices are compared. The temperature profile of the rest should be ok due to the previous timestep
                    # The while loop will check the two upper slices until the temperature profile is correct. This can tanke more loops because the secound slice is always replaced by the third after it is combined with the  first
                    while len(self.my_slices) > 1 and outer_slice.temperature < self.my_slices[1].temperature:
                        outer_slice.add_another_slice(self.my_slices[1])
                        self.my_slices.remove(self.my_slices[1])

            elif not is_from_top:
                outer_slice = self.my_slices[-1]
                if ws.temperature <= (outer_slice.temperature - temperature_difference):
                    self.my_slices.append(ws)
                else:
                    outer_slice.add_another_slice(ws)
                    # connect the bottom slices till the temperature profile is correct
                    while len(self.my_slices) > 1 and self.my_slices[-1].temperature > self.my_slices[-2].temperature:
                        self.my_slices[-2].add_another_slice(self.my_slices[-1])
                        self.my_slices.remove(self.my_slices[-1])

    # def push_slices(self, ws_height, is_from_top: bool):
    def push_slices(self, ws_height: Any, is_from_top: bool) -> Any:
        """
        While the current top or bottom slice has a height of zero it is not existing and has to be deleted.

        Initially an empty pushed_out_slice is created.
        When the system is not in operation, slices must still be created. Their size is zero and they are deleted in the next timestep.
        If the system is operating the same amount of water, that came into the tank (ws), has to be pushed out of the tank.
        Therefore the slices on the opposite side are collected and mixed with the empty slice.
        This way a new output slice is created and the collected slices are deleted.
        Its also possible to collect only a part of a slice. The rest will stay in the tank a the outer layer.

        Caution:
        slice_to_push =     the previously existing slice in the tank
        pushed_out_slice =  the new created slice which will leave the tank

        :param ws_height:   Height of the incoming and already inserted water slice. The same slice height has to be pushed out
        :param is_from_top: Where was the slice inserted --> water will be pushed on the opposite side
        :return:            The slice which is pushed out in this timestep on the opposite side of bool: is_from_top
        """
        # maybe redundant lines?
        while self.my_slices[0].height == 0:
            self.my_slices.remove(self.my_slices[0])
            raise ValueError
        while self.my_slices[-1].height == 0:
            self.my_slices.remove(self.my_slices[-1])
            raise ValueError

        pushed_out_slice = WaterSlice(self.diameter, 0, 0)
        if ws_height > 0:
            collected_height_so_far:float = 0
            while collected_height_so_far < ws_height:
                height_still_needed = ws_height - collected_height_so_far
                if is_from_top:
                    slice_to_push = self.my_slices[-1]  # The slice which will added to the pushed_out_slice next
                else:  # is_fom_top == False
                    slice_to_push = self.my_slices[0]
                if slice_to_push.height <= height_still_needed:         # the whole slice will be pushed
                    self.my_slices.remove(slice_to_push)                # removing whole slice
                    pushed_out_slice.add_another_slice(slice_to_push)   # add the whole slice to pushed_out_slice
                    collected_height_so_far += slice_to_push.height
                else:   # Only a part of the outermost layer has to be removed. This part will be added to pushed_out_slice and removed from the remaining slice
                    pushed_out_slice.add_another_slice(WaterSlice(self.diameter, height_still_needed, slice_to_push.temperature))
                    # Height, mass & enthalpy have to be changed in the remaining slice
                    slice_to_push.height -= height_still_needed
                    slice_to_push.mass -= (height_still_needed * self.area * pushed_out_slice.density)
                    slice_to_push.enthalpy -= (height_still_needed * self.area * pushed_out_slice.density) * slice_to_push.specific_heat_capacity * slice_to_push.temperature
                    collected_height_so_far += height_still_needed  # -> this should be zero
                    if collected_height_so_far != ws_height:
                        a = collected_height_so_far
                        b = ws_height


                    # The remaining slice (slice_to_push) must have a minimum size
                    if slice_to_push.height < WarmWaterStorageConfig.slice_height_minimum:
                        self.my_slices.remove(slice_to_push)
                        if is_from_top:
                            self.my_slices[-1].add_another_slice(slice_to_push)
                        else:
                            self.my_slices[0].add_another_slice(slice_to_push)
        return pushed_out_slice

    def check_slice_temperature_order(self) -> None:
        """
        Heat losses on top and bottom lead to a decrease of these slices.
        Especially if there are no changes in the tank (in-/output) this can lead to higher changes in the temperature of the top/bottom slice
        The temperature profile of the two highest and the two lowest slices are checked.
        The slices in between are not influenced by this effect.
        """

        # check from top
        if len(self.my_slices) >= 2:
            while self.my_slices[0].temperature <= self.my_slices[1].temperature:
                self.my_slices[0].add_another_slice(self.my_slices[1])
                self.my_slices.remove(self.my_slices[1])
                # no more mixing if there is only one slice left
                if len(self.my_slices) == 1:
                    break

        # check from bottom
        if len(self.my_slices) >= 2:
            while self.my_slices[-1].temperature >= self.my_slices[-2].temperature:
                self.my_slices[-2].add_another_slice(self.my_slices[-1])
                self.my_slices.remove(self.my_slices[-1])
                # no more mixing if there is only one slice left
                if len(self.my_slices) == 1:
                    break

    def simulate_one_timestep(self, seconds_per_timestep: int, slice_temperature_input_upper: float,
                              slice_mass_input_upper: float, is_from_top_upper: bool,
                              slice_temperature_input_bottom: float, slice_mass_input_bottom: float, is_from_top_bottom: bool) -> Any:
        """
        This is the combination of the previous functions.
        One timestep will be simulated. This includes 5 steps:
        1. Creation of the two new input slices according to the input values
        2. Insert the slices into the tank. This can include the mixing of slices inside the tank.
            -> It also leads to a 'larger' height of the tank for the moment.
        3. Create the output slices and delete the corresponding slices in the tank
            --> The tank has its initial size again (tank size == sum of all slices)
        4. Heat losses for all slices in the tank (horizontal). (Vertical losses (top/bottom) could be added later on)

        :param seconds_per_timestep:            seconds per timestep
        :param slice_temperature_input_upper:   temperature of water chp --> storage
        :param slice_mass_input_upper:          mass of water chp --> storage
        :param is_from_top_upper:               True value
        :param slice_temperature_input_bottom:  temperature of water load --> storage
        :param slice_mass_input_bottom:         mass of water load --> storage
        :param is_from_top_bottom:              False value

        :return:    output_top      water slices which is pushed out at the top
                    output_bottom   water slices which is pushed out at the bottom
                    heat_losses [W] heat losses in this timestep
        """
        # in and outflow
        ws_upper = self.create_water_slice(slice_temperature_input_upper, slice_mass_input_upper)
        ws_bottom = self.create_water_slice(slice_temperature_input_bottom, slice_mass_input_bottom)

        ws_upper_height = deepcopy(ws_upper.height)
        ws_bottom_height = deepcopy(ws_bottom.height)

        if ws_upper.height != slice_mass_input_upper * 4 / (pi * (self.diameter ** 2) * PhysicsConfig.water_density):
            log.information("huhuu")
            log.information(str(ws_upper.height))
            log.information(str(slice_mass_input_upper * 4 / (pi * (self.diameter ** 2) * PhysicsConfig.water_density)))
            raise ValueError

        if ws_bottom.height != slice_mass_input_bottom * 4 / (pi * (self.diameter ** 2) * PhysicsConfig.water_density):
            log.information("huhuu")
            log.information(str(ws_bottom.height))
            log.information(str(slice_mass_input_bottom * 4 / (pi * (self.diameter ** 2) * PhysicsConfig.water_density)))
            raise ValueError

        previous_mass = self.calculate_tanks_mass()

        # insert slices
        # CHP
        self.insert_slice(ws_upper, is_from_top_upper)
        # Load
        self.insert_slice(ws_bottom, is_from_top_bottom)

        # push slices
        # to Load
        output_top = self.push_slices(ws_bottom_height, is_from_top_bottom)
        # to CHP
        output_bottom = self.push_slices(ws_upper_height, is_from_top_upper)

        mass_now = self.calculate_tanks_mass()

        if abs(mass_now - previous_mass) > 0.001:
            raise ValueError

        # heat losses to environment and energy exchange between the segments
        if len(self.my_slices) > 1:
            self.energy_exchange_between_slices(lambda_water_water=0.6, seconds_per_timestep=seconds_per_timestep)

        # calculate heat losses
        heat_losses = self.energy_losses_in_one_timestep(WarmWaterStorageConfig.tank_u_value, seconds_per_timestep)

        return output_top, output_bottom, heat_losses

    def get_temperature_level_at_specific_height(self, height_of_interest: float) -> Any:
        """
        Get the temperature at a specific height in the tank. Height is measured from top to bottom.
        The height of the slices in the tank will be added to collected_height if larger than height_of_interest.
        The slice counter must be lower than the slices in the tank.
        The slice counter tells how many slices have been skipped to get to the right height in the tank.
        To set temperature_of_interest correctly the slice counter has to be reduced by one before the temperature is assigned.
        Except the height_of_interest is zero. Then the top slice [0] is taken.

        :param height_of_interest:          The height of the slice whose temperature is to be returned
        :return: temperature_of_interest    Temperature of interest [°C]
        """
        collected_height:float = 0
        slice_counter:int = 0
        slices_in_tank = len(self.my_slices)
        while slice_counter < slices_in_tank and height_of_interest > collected_height:
            collected_height += self.my_slices[slice_counter].height
            slice_counter += 1
        if height_of_interest == 0:
            temperature_of_interest = self.my_slices[0].temperature
        else:
            # numbers of my_slice start at 0, so 1 has to be subtracted
            temperature_of_interest = self.my_slices[slice_counter - 1].temperature

        return temperature_of_interest

    def get_load_percentage(self, minimum_temperature: float) -> Any:
        """
        --> not used
        How big is the part of the tank which is above a certain temperature.
        :param minimum_temperature: Only slices with this temperature or above this temperature-level can be used properly
        :return: usable_percentage_of_tank: % which are above minimum temperature
        """
        usable_height:float = 0
        slice_counter = 0
        slices_in_tank = len(self.my_slices)
        while slice_counter < slices_in_tank and self.my_slices[slice_counter].temperature >= minimum_temperature:
            usable_height += self.my_slices[slice_counter].height
            slice_counter += 1
        usable_percentage_of_tank = usable_height / self.height_storage * 100

        return usable_percentage_of_tank, usable_height

    def calculate_tanks_enthalpy(self) -> Any:
        """
        To get the enthalpy of all slices combined
        No changes of values
        Calculation from Ws to kWh; 3600s = 1h, 1000W = 1kW
        :return enthalpy_tank   Enthalpy of the tank [kWh]
        """
        enthalpy_tank:float = 0
        for i in range(len(self.my_slices)):
            enthalpy_tank += self.my_slices[i].enthalpy

        enthalpy_tank = enthalpy_tank / 3600 / 1000
        return enthalpy_tank

    def calculate_tanks_mean_temperature(self) -> Any:
        """
        Get the average temperature of tank
        specific_heat_capacity can be neglected because its constant for all slices
        For each slice:
        temperature(i) * mass(i) / total_mass
        :return average_temperature     Average tank temperature [°C]
        """
        total_energy:float = 0
        average_density = PhysicsConfig.water_density
        for i in range(len(self.my_slices)):
            total_energy += self.my_slices[i].mass * self.my_slices[i].temperature
        average_temperature:float = total_energy / (self.volume * average_density)
        return average_temperature

    def calculate_tanks_mass(self) -> Any:
        total_mass:float = 0
        for i in range(len(self.my_slices)):
            total_mass += self.my_slices[i].mass
        return total_mass

    def energy_losses_in_one_timestep(self, u_value_tank: Any, seconds_per_timestep: int = 1, ambient_temperature: float = 20.0) -> Any:
        """
        Energy losses for all the slices in the tank
        The called function 'heat_losses_to_ambient' & 'heat_losses_top_bottom' sets a new enthalpy and temperature for slices
        :return losses_this_timestep [Ws]    Sum of all heat losses to the ambient in this timestep
        """
        losses_this_timestep = 0.0
        # Losses at top and bottom
        losses_this_timestep += self.my_slices[0].heat_losses_vertical_top_or_bottom(u_value_tank, seconds_per_timestep, ambient_temperature)
        losses_this_timestep += self.my_slices[-1].heat_losses_vertical_top_or_bottom(u_value_tank, seconds_per_timestep, ambient_temperature)

        # Losses to the sides
        for i in range(len(self.my_slices)):
            losses_this_timestep += self.my_slices[i].heat_losses_horizontal(u_value_tank, seconds_per_timestep, ambient_temperature)
        return losses_this_timestep

    def energy_exchange_between_slices(self, lambda_water_water: float = 0.6, seconds_per_timestep: int = 1) -> None:
        """
        Internal heat exchange between the slices
        The transferred heat is defined by the temperature difference and the distance between the segments middle.
        slice_height saves all the distances between the slices.
        energy_transfer saves the energy which is transferred to the slice below.
        The transferred energy ( = Enthalpy) is subtracted from the slices and a new temperature is calculated.
        The storing of transferred energy is necessary because a direct subtraction would influence the iterative process.

        u_value Water-Water = ~0.6 ?!
        Upper slice (warmer):
            energy is transferred to the lower slice
        Upper slice (colder):
            energy is received from the upper slice

        --> top slice just looses energy
        --> bottom slice just gains energy
        """

        energy_transfer = []
        slice_height = []
        # for i in range(len(self.my_slices)):
            # log.information("Schicht: " + str(i) + " Höhe: " + str(self.my_slices[i].height) + " Temperatur: " + str(self.my_slices[i].temperature))

        for i in range(len(self.my_slices) - 1):
            slice_height.append((self.my_slices[i].height + self.my_slices[i+1].height) / 2)

        for i in range(len(self.my_slices) - 1):
            energy_transfer.append((lambda_water_water * self.area / slice_height[i] * (self.my_slices[i].temperature - self.my_slices[i+1].temperature)) * seconds_per_timestep)

        assert len(energy_transfer) == len(slice_height)

        for i in range(len(self.my_slices) - 1):
            self.my_slices[i].enthalpy -= energy_transfer[i]
            self.my_slices[i+1].enthalpy += energy_transfer[i]

        for i in range(len(self.my_slices)):
            # log.information("slice number " + str(i))
            previous = deepcopy(self.my_slices[i].temperature)
            self.my_slices[i].calculate_new_temperature()
            # log.information(previous)
            # log.information(self.my_slices[i].temperature)
            assert abs(previous - self.my_slices[i].temperature) > 0

    def energy_exchange_between_slices_differential_equation(self, lambda_water_water: float = 0.06, seconds_per_timestep: int = 1) -> None:
        """
        A differential equation can be used to solve this.
        Most complex part of the whole simulation.. is this even useful?
        ToDo: This code is not working so far.. Function above should be good enough

        ToDo: Add heat transfer to the environment

        Using odeint function from scipy.integrate
        :param lambda_water_water:
        :param seconds_per_timestep:
        :return:
        """
        slice_height = []
        for i in range(len(self.my_slices)):
            log.information("Schicht: " + str(i) + " Höhe: " + str(self.my_slices[i].height) + " Temperatur: " + str(self.my_slices[i].temperature))

        for i in range(len(self.my_slices) - 1):
            slice_height.append((self.my_slices[i].height + self.my_slices[i+1].height) / 2)
        slice_height.append(0)

        def ode_sd():
            # ab 2. Schicht
            division_factor = (self.my_slices[i].specific_heat_capacity * self.my_slices[i].mass)
            coefficient_T_slice = ((lambda_water_water * self.my_slices[i].area) / slice_height[i] + (lambda_water_water * self.my_slices[i].area)/ slice_height[i+1]) / division_factor

            coefficient_constant_T_slice = ((lambda_water_water * self.my_slices[i].area * self.my_slices[i-1].temperature) / slice_height[i] + (lambda_water_water * self.my_slices[i].area * self.my_slices[i+1].temperature)/ slice_height[i+1]) / division_factor

            dT_waterslice_dt = (-1) * self.my_slices[i].temperature * coefficient_T_slice + coefficient_constant_T_slice

            return dT_waterslice_dt
        new_temperatures:List[float] = []
        for i in range(len(self.my_slices)):
            # y = odeint(ode_sd, y0, ts)
            # new_temperatures.append(y)
            pass
        for i in range(len(self.my_slices)):
            self.my_slices[i].temperature = new_temperatures[i]
        pass


class StratifiedWarmWaterStorage(Component):

    # Water intake coming from the CHP
    CHP_ChargingSideInput_mass = "ChargingSideInput_mass"                       # kg/s
    CHP_ChargingSideInput_temperature = "ChargingSideInput_temperature"         # °C
    Heating_DischargingSideInput_mass = "Heating DischargingSideInput_mass"                 # kg/s
    Heating_DischargingSideInput_temperature = "Heating DischargingSideInput_temperature"   # °C

    # Water intake coming from the gas heater
    Gas_ChargingSideInput_mass = "Gas ChargingSideInput_mass"               # kg/s
    Gas_ChargingSideInput_temperature = "Gas ChargingSideInput_temperature" # °C

    # ww
    WW_DischargingSideInput_mass = "WW DischargingSideInput_mass"                 # kg/s
    WW_DischargingSideInput_temperature = "WW DischargingSideInput_temperature"   # °C

    # Water outtake going back to the CHP
    CHP_ChargingSideOutput_mass = "CHP ChargingSideOutput_mass"                       # kg/s
    CHP_ChargingSideOutput_temperature = "CHP ChargingSideOutput_temperature"         # °C
    Heating_DischargingSideOutput_mass = "Heating DischargingSideOutput_mass"               # kg/s
    Heating_DischargingSideOutput_temperature = "Heating DischargingSideOutput_temperature" # °C

    # Water outtake going back to the gas heater
    Gas_ChargingSideOutput_mass = "Gas ChargingSideOutput_mass"               # kg/s
    Gas_ChargingSideOutput_temperature = "Gas ChargingSideOutput_temperature" # °C

    # ww
    WW_DischargingSideOutput_mass = "WW DischargingSideOutput_mass"  # kg/s
    WW_DischargingSideOutput_temperature = "WW DischargingSideOutput_temperature"  # °C

    # temperatures
    Temperature0Percent = "Temperature 0 Percent"                           # °C
    Temperature20Percent = "Temperature 20 Percent"                         # °C
    Temperature40Percent = "Temperature 40 Percent"                         # °C
    Temperature60Percent = "Temperature 60 Percent"                         # °C
    Temperature80Percent = "Temperature 80 Percent"                         # °C
    Temperature100Percent = "Temperature 100 Percent"                       # °C

    # For information
    TankEnthalpy = "Tank_Enthalpy"                                          # kWh
    TankMeanTemperature = "Tank_Mean_Temperature"                           # °C
    AmountOfSlices = "Amount Of Slices"                                     # -
    HeatLosses = "Heat Losses"                                              # W
    TankMass = "Tank Mass"                                                  # kg

    def __init__(self, component_name: str,my_simulation_parameters: SimulationParameters,  config: WarmWaterStorageConfig) -> None:
        super().__init__(name=component_name, my_simulation_parameters=my_simulation_parameters)

        # Input
        self.chp_charging_side_input_mass: ComponentInput = self.add_input(self.component_name, StratifiedWarmWaterStorage.CHP_ChargingSideInput_mass, lt.LoadTypes.WARM_WATER, lt.Units.KG_PER_SEC, True)
        self.chp_charging_side_input_temperature: ComponentInput = self.add_input(self.component_name, StratifiedWarmWaterStorage.CHP_ChargingSideInput_temperature, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS, True)
        self.heating_discharging_side_input_mass: ComponentInput = self.add_input(self.component_name, StratifiedWarmWaterStorage.Heating_DischargingSideInput_mass, lt.LoadTypes.WARM_WATER, lt.Units.KG_PER_SEC, True)
        self.heating_discharging_side_input_temperature: ComponentInput = self.add_input(self.component_name, StratifiedWarmWaterStorage.Heating_DischargingSideInput_temperature, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS, True)
        # gas --> ?not mandatory?
        self.gas_charging_side_input_mass: ComponentInput = self.add_input(self.component_name, StratifiedWarmWaterStorage.Gas_ChargingSideInput_mass, lt.LoadTypes.WARM_WATER, lt.Units.KG_PER_SEC, True)
        self.gas_charging_side_input_temperature: ComponentInput = self.add_input(self.component_name, StratifiedWarmWaterStorage.Gas_ChargingSideInput_temperature, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS, True)
        # warm water
        self.ww_discharging_side_input_mass: ComponentInput = self.add_input(self.component_name, StratifiedWarmWaterStorage.WW_DischargingSideInput_mass, lt.LoadTypes.WARM_WATER, lt.Units.KG_PER_SEC, True)
        self.ww_discharging_side_input_temperature: ComponentInput = self.add_input(self.component_name, StratifiedWarmWaterStorage.WW_DischargingSideInput_temperature, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS, True)

        # Output
        self.chp_charging_side_output_mass: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.CHP_ChargingSideOutput_mass, lt.LoadTypes.WARM_WATER, lt.Units.KG_PER_SEC)
        self.chp_charging_side_output_temperature: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.CHP_ChargingSideOutput_temperature, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS)
        self.heating_discharging_side_output_mass: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.Heating_DischargingSideOutput_mass, lt.LoadTypes.WARM_WATER, lt.Units.KG_PER_SEC)
        self.heating_discharging_side_output_temperature: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.Heating_DischargingSideOutput_temperature, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS)
        # gas
        self.gas_charging_side_output_mass: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.Gas_ChargingSideOutput_mass, lt.LoadTypes.WARM_WATER, lt.Units.KG_PER_SEC)
        self.gas_charging_side_output_temperature: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.Gas_ChargingSideOutput_temperature, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS)
        # warm water
        self.ww_discharging_side_output_mass: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.WW_DischargingSideOutput_mass, lt.LoadTypes.WARM_WATER, lt.Units.KG_PER_SEC)
        self.ww_discharging_side_output_temperature: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.WW_DischargingSideOutput_temperature, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS)

        # temperatures
        self.temperature_0_percent: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.Temperature0Percent, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS)
        self.temperature_20_percent: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.Temperature20Percent, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS)
        self.temperature_40_percent: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.Temperature40Percent, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS)
        self.temperature_60_percent: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.Temperature60Percent, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS)
        self.temperature_80_percent: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.Temperature80Percent, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS)
        self.temperature_100_percent: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.Temperature100Percent, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS)

        # Outputs for information
        self.tank_enthalpy: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.TankEnthalpy, lt.LoadTypes.WARM_WATER, lt.Units.KWH)
        self.tank_mean_temperature: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.TankMeanTemperature, lt.LoadTypes.WARM_WATER, lt.Units.CELSIUS)
        self.amount_of_slices: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.AmountOfSlices, lt.LoadTypes.WARM_WATER, lt.Units.ANY)
        self.heat_losses: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.HeatLosses, lt.LoadTypes.WARM_WATER, lt.Units.WATT)
        self.tank_mass: ComponentOutput = self.add_output(self.component_name, StratifiedWarmWaterStorage.TankMass, lt.LoadTypes.WARM_WATER, lt.Units.KG)

        self.wws = WarmWaterStorageSimulation(config)
        self.seconds_per_timestep = my_simulation_parameters.seconds_per_timestep

        self.previous_state = 0

    def i_save_state(self) -> None:
        self.previous_state = self.wws.begin_new_timestep()

    def i_restore_state(self) -> None:
        self.wws.reset_to_last_timestep(self.previous_state)

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:

        # temperature of the tank at different heights
        # 0 = top, 100 = bottom
        height_one_percent = WarmWaterStorageConfig.tank_height / 100
        temperature_0_percent = self.wws.get_temperature_level_at_specific_height(height_one_percent * 0)
        temperature_20_percent = self.wws.get_temperature_level_at_specific_height(height_one_percent * 20)
        temperature_40_percent = self.wws.get_temperature_level_at_specific_height(height_one_percent * 40)
        temperature_60_percent = self.wws.get_temperature_level_at_specific_height(height_one_percent * 60)
        temperature_80_percent = self.wws.get_temperature_level_at_specific_height(height_one_percent * 80)
        temperature_100_percent = self.wws.get_temperature_level_at_specific_height(height_one_percent * 100)

        tank_enthalpy = self.wws.calculate_tanks_enthalpy()
        tank_mean_temperature = self.wws.calculate_tanks_mean_temperature()
        # sometimes there occur rounding errors at the ~10th decimal digit. they can be neglected
        tank_mass = self.wws.calculate_tanks_mass()

        # amount of slices in the tank
        amount_of_slices = len(self.wws.my_slices)

        # water slice input values
        # mass is in kg/s --> converted to kg/timestep

        # charging side
        # from CHP
        chp_charging_side_input_mass_sec = stsv.get_input_value(self.chp_charging_side_input_mass)
        chp_charging_side_input_mass = chp_charging_side_input_mass_sec * self.seconds_per_timestep
        chp_charging_side_input_temperature = stsv.get_input_value(self.chp_charging_side_input_temperature)
        # from gas
        gas_charging_side_input_mass_sec = stsv.get_input_value(self.gas_charging_side_input_mass)
        gas_charging_side_input_mass = gas_charging_side_input_mass_sec * self.seconds_per_timestep
        gas_charging_side_input_temperature = stsv.get_input_value(self.gas_charging_side_input_temperature)
        # total charging input
        total_charging_side_input_mass = chp_charging_side_input_mass + gas_charging_side_input_mass
        is_from_top_upper = True

        if total_charging_side_input_mass > 0:
            total_charging_side_input_temperature = (chp_charging_side_input_mass * chp_charging_side_input_temperature + gas_charging_side_input_mass * gas_charging_side_input_temperature) / total_charging_side_input_mass
        else:
            # fake value for empty slice -> 99.123 --> no CHP and no gas-heater active
            total_charging_side_input_temperature = 99.123

        # discharging side
        # from heating
        heating_discharging_side_input_mass_sec = stsv.get_input_value(self.heating_discharging_side_input_mass)
        heating_discharging_side_input_mass = heating_discharging_side_input_mass_sec * self.seconds_per_timestep
        heating_discharging_side_input_temperature = stsv.get_input_value(self.heating_discharging_side_input_temperature)
        # from ww
        ww_discharging_side_input_mass_sec = stsv.get_input_value(self.ww_discharging_side_input_mass)
        ww_discharging_side_input_mass = ww_discharging_side_input_mass_sec * self.seconds_per_timestep
        ww_discharging_side_input_temperature = stsv.get_input_value(self.ww_discharging_side_input_temperature)

        # total discharging input
        total_discharging_side_input_mass = heating_discharging_side_input_mass + ww_discharging_side_input_mass
        is_from_top_bottom = False

        if total_discharging_side_input_mass > 0:
            total_discharging_side_input_temperature = (heating_discharging_side_input_mass * heating_discharging_side_input_temperature + ww_discharging_side_input_mass * ww_discharging_side_input_temperature) / total_discharging_side_input_mass
        else:
            # fake value for empty slice -> 66.123
            total_discharging_side_input_temperature = 66.123

        # actual simulation of wws
        output_top, output_bottom, heat_losses = self.wws.simulate_one_timestep(self.seconds_per_timestep,
                                                                                total_charging_side_input_temperature,
                                                                                total_charging_side_input_mass,
                                                                                is_from_top_upper,
                                                                                total_discharging_side_input_temperature,
                                                                                total_discharging_side_input_mass,
                                                                                is_from_top_bottom)

        # Check temperature of the two highest and the two lowest slices
        self.wws.check_slice_temperature_order()

        # Ws --> W
        heat_losses_watt = heat_losses / self.seconds_per_timestep

        # set outputs

        # water incompressibility -> what mass comes in from one side, has to go out at the same side
        # massflows are equal to the inflows
        stsv.set_output_value(self.chp_charging_side_output_mass, chp_charging_side_input_mass_sec)
        stsv.set_output_value(self.gas_charging_side_output_mass, gas_charging_side_input_mass_sec)
        stsv.set_output_value(self.heating_discharging_side_output_mass, heating_discharging_side_input_mass_sec)
        stsv.set_output_value(self.ww_discharging_side_output_mass, ww_discharging_side_input_mass_sec)

        # output temperatures
        stsv.set_output_value(self.chp_charging_side_output_temperature, output_bottom.temperature)
        stsv.set_output_value(self.gas_charging_side_output_temperature, output_bottom.temperature)
        stsv.set_output_value(self.heating_discharging_side_output_temperature, output_top.temperature)
        stsv.set_output_value(self.ww_discharging_side_output_temperature, output_top.temperature)

        # measured temperatures
        stsv.set_output_value(self.temperature_0_percent, temperature_0_percent)
        stsv.set_output_value(self.temperature_20_percent, temperature_20_percent)
        stsv.set_output_value(self.temperature_40_percent, temperature_40_percent)
        stsv.set_output_value(self.temperature_60_percent, temperature_60_percent)
        stsv.set_output_value(self.temperature_80_percent, temperature_80_percent)
        stsv.set_output_value(self.temperature_100_percent, temperature_100_percent)

        # for information
        stsv.set_output_value(self.tank_enthalpy, tank_enthalpy)
        stsv.set_output_value(self.tank_mean_temperature, tank_mean_temperature)
        stsv.set_output_value(self.amount_of_slices, amount_of_slices)
        stsv.set_output_value(self.heat_losses, heat_losses_watt)
        stsv.set_output_value(self.tank_mass, tank_mass)

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        # alle ausgabewerte die zu überprüfen sind können hiermit fehlerausgabeüberprüft werden
        pass



""" Simple hot water storage implementation.

Energy bucket model: extracts energy, adds energy and converts back to temperatere.
The hot water storage simulates only storage and demand and needs to be connnected to a heat source. It can act as boiler with input:
hot water demand or as  buffer with input ThermalPowerToBuilding. Both options need input signal for heating power and have
one output: the hot water storage temperature.
"""
# clean
# Generic/Built-in
from typing import Optional, List, Any
from dataclasses import dataclass
from dataclasses_json import dataclass_json

import numpy as np

# Owned
import hisim.component as cp
from hisim.components.loadprofilegenerator_utsp_connector import UtspLpgConnector
import hisim.dynamic_component as dycp
from hisim import loadtypes as lt
from hisim.simulationparameters import SimulationParameters
from hisim.components.loadprofilegenerator_connector import Occupancy
from hisim.components import generic_heat_pump_modular
from hisim.components import generic_heat_source
from hisim.components import controller_l1_generic_runtime
import hisim.log

__authors__ = "Johanna Ganglbauer - johanna.ganglbauer@4wardenergy.at"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = "MIT"
__version__ = "0.1"
__maintainer__ = "Vitor Hugo Bellotto Zago"
__email__ = "vitor.zago@rwth-aachen.de"
__status__ = "development"


@dataclass_json
@dataclass
class StorageConfig:

    """ Used in the HotWaterStorageClass defining the basics. """

    name: str
    use: lt.ComponentType
    source_weight: int
    volume: float
    surface: float
    u_value: float
    warm_water_temperature: float
    drain_water_temperature: float
    efficiency: float
    power: Optional[float]
    cooling_considered: bool
    heating_season_begin: Optional[int]
    heating_season_end: Optional[int]

    def __init__(self, name: str, use: lt.ComponentType, source_weight: int, volume: float, surface: float,
                 u_value: float, warm_water_temperature: float, drain_water_temperature: float, efficiency: float,
                 power: Optional[float], cooling_considered: bool, heating_season_begin: Optional[int],
                 heating_season_end: Optional[int]):
        """ Initializes Storage Config. """
        self.name = name
        self.use = use
        self.source_weight = source_weight
        self.volume = volume
        self.surface = surface
        self.u_value = u_value
        self.warm_water_temperature = warm_water_temperature + 273.15
        self.drain_water_temperature = drain_water_temperature + 273.15
        self.efficiency = efficiency
        self.power = power
        self.cooling_considered = cooling_considered
        self.heating_season_begin = heating_season_begin
        self.heating_season_end = heating_season_end

    @staticmethod
    def get_default_config_boiler():
        """ Returns default configuration for boiler. """
        config = StorageConfig(name='DHWBoiler', use=lt.ComponentType.BOILER, source_weight=1, volume=500,
                               surface=2.0, u_value=0.36, warm_water_temperature=50, drain_water_temperature=10,
                               efficiency=1, power=None, cooling_considered=False, heating_season_begin=None,
                               heating_season_end=None)
        return config

    @staticmethod
    def get_default_config_buffer(volume: float = 500) -> Any:
        """ Returns default configuration for buffer (radius:height = 1:4). """
        radius = (volume * 1e-3 / (4 * np.pi))**(1 / 3)
        config = StorageConfig(
            name='Buffer', use=lt.ComponentType.BUFFER, source_weight=1, volume=volume, surface=6 * radius * radius * np.pi, u_value=0.36,
            warm_water_temperature=50, drain_water_temperature=10, efficiency=1, power=1500, cooling_considered=True,
            heating_season_begin=270, heating_season_end=150)
        return config


class StorageState:

    """ Data class saves the state of the simulation results. """

    def __init__(self, timestep: int = -1, volume_in_l: float = 200, temperature_in_kelvin: float = 273.15 + 50):
        """Initializes instance of class Storage State.

        Parameters
        ----------
        timestep: int, optional
            Timestep of simulation. The default is 0.
        volume_in_l: float
            Volume of hot water storage in liters.
        temperature_in_kelvin: float
            Temperature of hot water storage in Kelvin.

        """
        self.timestep = timestep
        self.temperature_in_kelvin = temperature_in_kelvin
        self.volume_in_l = volume_in_l

    def clone(self):
        """ Replicates storage state. """
        return StorageState(self.timestep, self.volume_in_l, self.temperature_in_kelvin)

    def energy_from_temperature(self) -> float:
        """ Converts temperature of storage (K) into energy contained in storage (kJ). """
        # 0.977 is the density of water in kg/l
        # 4.182 is the specific heat of water in kJ / (K * kg)
        return self.temperature_in_kelvin * self.volume_in_l * 0.977 * 4.182  # energy given in kJ

    def set_temperature_from_energy(self, energy_in_kilo_joule):
        """ Converts energy contained in storage (kJ) into temperature (K). """
        # 0.977 is the density of water in kg/l
        # 4.182 is the specific heat of water in kJ / (K * kg)
        self.temperature_in_kelvin = energy_in_kilo_joule / (self.volume_in_l * 0.977 * 4.182)  # temperature given in K
        # filter for boiling water
        # no filtering -> this hides major problems - Noah
        if self.temperature_in_kelvin > 95 + 273.15:
            raise ValueError("Water was boiling. This points towards a major problem in your model.")
        # filter for freezing water
        if self.temperature_in_kelvin < 2 + 273.15:
            raise ValueError("Water in your storage tank was freezing. This points towards a major problem in your model.")

    def return_available_energy(self, heating: bool) -> float:
        """ Returns available energy in (J).

        For heating up the building in winter and cooling the building in summer in case of buffer storages.
        Here 30°C is set as the lower limit for the temperature in the buffer storage in winter and 20°C
        is set as the upper limit for the temperature in buffer for cooling in summer.
        """
        if heating:
            available_energy = (self.temperature_in_kelvin - 273.15 - 25) * self.volume_in_l * 0.977 * 4.182 * 1e3
        else:
            available_energy = (self.temperature_in_kelvin - 273.15 - 21) * self.volume_in_l * 0.977 * 4.182 * 1e3
        return available_energy


class HotWaterStorage(dycp.DynamicComponent):

    """Simple hot water storage implementation.

    Energy bucket model: extracts energy, adds energy and converts back to temperatere.
    The hot water storage simulates only storage and demand and needs to be connnected to a heat source. It can act as boiler with input:
    hot water demand or as  buffer with input ThermalPowerToBuilding. Both options need input signal for heating power and have
    one output: the hot water storage temperature.

    Parameters
    ----------
    name: str, optional
        Name of hot water storage within simulation. The default is 'Boiler'.
    use: lt.ComponentType, optional
        Use of hot water storage either as Boiler or as Buffer
    source_weight: int, optional
        Weight of component, relevant if there is more than one hot water storage, defines hierachy in control. The default is 1.
    volume: float, optional
        Volume of storage in liters. The default is 200 l.
    surface: float, optional
        Surface of storage in square meters. The default is 1.2 m**3
    u_value: float, optional
        u-value of stroage in W m**(-2) K**(-1). The default is 0.36 W / (m*m*K)
    warm_water_temperature: float, optional
        Set temperature of hot water used by residents in °C. The default is 50 °C.
    drain_water_temperature: float, optional
        Temperature of cold water from pipe in °C. The default is 10 °C.
    efficiency: float, optional
        Efficiency of the heating rod / hydrogen oven. The default is 1.

    """

    # Inputs
    ThermalPowerDelivered = "ThermalPowerDelivered"  # either thermal energy delivered
    WaterConsumption = "WaterConsumption"
    L1DeviceSignal = "L1DeviceSignal"
    my_component_inputs: List[dycp.DynamicConnectionInput] = []
    my_component_outputs: List[dycp.DynamicConnectionOutput] = []

    # obligatory Outputs
    TemperatureMean = "TemperatureMean"

    # outputs for buffer storage
    HeatToBuilding = "HeatToBuilding"

    def __init__(self, my_simulation_parameters: SimulationParameters, config: StorageConfig):
        """ Initializes instance of HotWaterStorage class. """

        super().__init__(my_component_inputs=self.my_component_inputs, my_component_outputs=self.my_component_outputs,
                         name=config.name + '_w' + str(config.source_weight), my_simulation_parameters=my_simulation_parameters)

        self.build(config)

        # initialize Boiler State
        self.state = StorageState(volume_in_l=config.volume, temperature_in_kelvin=273 + 60)
        self.previous_state = self.state.clone()
        self.write_to_report()

        # inputs
        if self.use == lt.ComponentType.BOILER:
            self.water_consumption_c: cp.ComponentInput = self.add_input(self.component_name, self.WaterConsumption, lt.LoadTypes.WARM_WATER,
                                                                         lt.Units.LITER, mandatory=True)
            self.add_default_connections(Occupancy, self.get_occupancy_default_connections())
            self.add_default_connections(UtspLpgConnector, self.get_utsp_default_connections())
        elif self.use == lt.ComponentType.BUFFER:
            self.l1_device_signal_c: cp.ComponentInput = self.add_input(self.component_name, self.L1DeviceSignal, lt.LoadTypes.ON_OFF,
                                                                        lt.Units.BINARY, mandatory=True)
            self.add_default_connections(controller_l1_generic_runtime.L1GenericRuntimeController, self.get_l1_default_connections())
        else:
            hisim.log.error('Type of hot water storage is not defined')

        self.thermal_power_delivered_c: cp.ComponentInput = self.add_input(self.component_name, self.ThermalPowerDelivered,
                                                                           lt.LoadTypes.HEATING, lt.Units.WATT, mandatory=False)

        # Outputs
        self.temperature_mean_c: cp.ComponentOutput = self.add_output(
            object_name=self.component_name, field_name=self.TemperatureMean, load_type=lt.LoadTypes.TEMPERATURE,
            unit=lt.Units.CELSIUS, postprocessing_flag=[lt.InandOutputType.STORAGE_CONTENT])
        # Outputs
        self.heat_to_building_c: cp.ComponentOutput = self.add_output(self.component_name, self.HeatToBuilding,
                                                                      lt.LoadTypes.HEATING, lt.Units.WATT)

        self.add_default_connections(generic_heat_pump_modular.ModularHeatPump, self.get_heatpump_default_connections())
        self.add_default_connections(generic_heat_source.HeatSource, self.get_heatpump_default_connections())

    def get_occupancy_default_connections(self):
        """ Sets occupancy default connections in hot water storage. """
        hisim.log.information("setting occupancy default connections in hot water storage")
        connections = []
        occupancy_classname = Occupancy.get_classname()
        connections.append(cp.ComponentConnection(HotWaterStorage.WaterConsumption, occupancy_classname,
                                                  Occupancy.WaterConsumption))
        return connections

    def get_utsp_default_connections(self):
        """ Sets occupancy default connections in hot water storage. """
        hisim.log.information("setting utsp default connections in hot water storage")
        connections = []
        utsp_classname = Occupancy.get_classname()
        connections.append(cp.ComponentConnection(HotWaterStorage.WaterConsumption, utsp_classname,
                                                  UtspLpgConnector.WaterConsumption))
        return connections

    def get_l1_default_connections(self):
        """ Sets L1 power default connections in hot water storage. """
        hisim.log.information("setting L1 power default connections in hot water storage")
        connections = []
        l1_classname = controller_l1_generic_runtime.L1GenericRuntimeController.get_classname()
        connections.append(cp.ComponentConnection(HotWaterStorage.L1DeviceSignal, l1_classname,
                                                  controller_l1_generic_runtime.L1GenericRuntimeController.L1DeviceSignal))
        return connections

    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass

    def get_heatpump_default_connections(self):
        """ Sets heat pump default connections in hot water storage. """
        hisim.log.information("setting heat pump default connections in hot water storage")
        connections = []
        heatpump_classname = generic_heat_pump_modular.ModularHeatPump.get_classname()
        connections.append(cp.ComponentConnection(HotWaterStorage.ThermalPowerDelivered, heatpump_classname,
                                                  generic_heat_pump_modular.ModularHeatPump.ThermalPowerDelivered))
        return connections

    def get_heatsource_default_connections(self):
        """ Sets heat source default connections in hot water storage. """
        hisim.log.information("setting heat source default connections in hot water storaage")
        connections = []
        heatsource_classname = generic_heat_source.HeatSource.get_classname()
        connections.append(cp.ComponentConnection(HotWaterStorage.ThermalPowerDelivered, heatsource_classname,
                                                  generic_heat_source.HeatSource.ThermalPowerDelivered))
        return connections

    def build(self, config: StorageConfig) -> None:
        """ Initializes hot water storage instance. """

        self.name = config.name
        self.use = config.use
        self.source_weight = config.source_weight
        self.volume = config.volume
        self.surface = config.surface
        self.u_value = config.u_value
        self.efficiency = config.efficiency
        self.drain_water_temperature = config.drain_water_temperature
        self.warm_water_temperature = config.warm_water_temperature
        self.power = config.power
        self.cooling_considered = config.cooling_considered
        if self.cooling_considered:
            if isinstance(config.heating_season_begin, int) and isinstance(config.heating_season_end, int):
                self.heating_season_begin = config.heating_season_begin * 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep
                self.heating_season_end = config.heating_season_end * 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep
            else:
                raise Exception("Heating season needs to be defined in config if cooling is considered")

    def write_to_report(self):
        """ Writes to report. """
        lines = []
        lines.append(f"Name: {self.name + str(self.source_weight)}")
        lines.append(f"Volume: {self.volume:4.0f} l")
        return lines

    def i_save_state(self):
        """ Abstract. Gets called at the beginning of a timestep to save the state. """
        self.previous_state = self.state.clone()

    def i_restore_state(self):
        """ Abstract. Restores the state of the component. Can be called many times while iterating. """
        self.state = self.previous_state.clone()

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """ Simulates iteration of hot water storage. """

        if self.thermal_power_delivered_c.source_output is not None:
            thermal_power_delivered = stsv.get_input_value(self.thermal_power_delivered_c)\
                * self.my_simulation_parameters.seconds_per_timestep * 1e-3  # 1e-3 conversion J to kJ
        else:
            thermal_power_delivered = sum(self.get_dynamic_inputs(stsv=stsv, tags=[lt.InandOutputType.HEAT_TO_BUFFER])) \
                * self.my_simulation_parameters.seconds_per_timestep * 1e-3  # 1e-3 conversion J to kJ
        heatconsumption: float = self.calculate_heat_consumption(stsv, thermal_power_delivered, timestep)
        stsv.set_output_value(self.heat_to_building_c, heatconsumption)

        # constant heat loss of heat storage with the assumption that environment has 20°C = 293 K -> based on energy balance in kJ
        # heat gain due to heating of storage -> based on energy balance in kJ
        energy = self.state.energy_from_temperature()
        new_energy = energy - (self.state.temperature_in_kelvin - 293) * self.surface * self.u_value *\
            self.my_simulation_parameters.seconds_per_timestep * 1e-3 - heatconsumption + thermal_power_delivered

        # convert new energy to new temperature
        self.state.set_temperature_from_energy(new_energy)

        # save outputs
        stsv.set_output_value(self.temperature_mean_c, self.state.temperature_in_kelvin - 273.15)

    def calculate_heat_consumption(self, stsv: cp.SingleTimeStepValues, thermal_power_delivered: float, timestep: int) -> float:
        """ Calculates the heat consumption. """
        if self.use == lt.ComponentType.BOILER:
            # heat loss due to hot water consumption -> base on energy balance in kJ
            # 0.977 density of water in kg/l
            # 4.182 specific heat of water in kJ K^(-1) kg^(-1)
            return stsv.get_input_value(self.water_consumption_c) \
                * (self.warm_water_temperature - self.drain_water_temperature) * 0.977 * 4.182
        if self.use == lt.ComponentType.BUFFER:
            if not isinstance(self.power, float):
                raise Exception("Power for heat transfer from buffer to building needs to be defined.")
            heatconsumption = stsv.get_input_value(self.l1_device_signal_c) \
                * self.power * self.my_simulation_parameters.seconds_per_timestep * 1e-3  # 1e-3 conversion J to kJ
            if self.cooling_considered:
                if self.heating_season_end < timestep < self.heating_season_begin:
                    available_power = self.state.return_available_energy(heating=False) \
                        / self.my_simulation_parameters.seconds_per_timestep + thermal_power_delivered
                    heatconsumption = -heatconsumption
                    if heatconsumption < available_power:
                        heatconsumption = min(available_power, 0)
                    return heatconsumption

                available_power = self.state.return_available_energy(
                    heating=True) / self.my_simulation_parameters.seconds_per_timestep + thermal_power_delivered
                if heatconsumption > available_power:
                    heatconsumption = max(available_power, 0)
            else:
                available_power = self.state.return_available_energy(
                    heating=True) / self.my_simulation_parameters.seconds_per_timestep + thermal_power_delivered
                if heatconsumption > available_power:
                    heatconsumption = max(available_power, 0)
        return heatconsumption



"""PREPROCESSING CALCULATES DIFFERENT SYSTEM VARIABLES E.G. INVESTMENT COSTS."""
from os import path
import json
import scipy.interpolate
import hisim.log
import hisim.utils


def calculate_pv_investment_cost(economic_parameters, pv_included, pv_peak_power):
    """PV."""
    if economic_parameters["pv_bought"]:
        if not pv_included:
            hisim.log.information("Error: PV bought but not included")
        else:
            ccpv = json.load(open(file=path.join(hisim.utils.HISIMPATH['modular_household'], 'ComponentCostPV.json'), mode='r'))
            pv_cost = scipy.interpolate.interp1d(ccpv["capacity_for_cost"], ccpv["cost_per_capacity"])
            pv_cost = pv_cost(pv_peak_power)
    else:
        pv_cost = 0
    return pv_cost


def calculate_smart_devices_investment_cost(economic_parameters, smart_devices_included):
    """SMART DEVICES."""
    if economic_parameters["smart_devices_bought"]:
        if not smart_devices_included:
            hisim.log.information("Error: Smart Devices bought but not included")
        else:
            ccsd = json.load(open(file=path.join(hisim.utils.HISIMPATH['modular_household'], 'ComponentCostSmartDevice.json'), mode='r'))
            smart_devices_cost = ccsd["smart_devices_cost"]
    else:
        smart_devices_cost = 0
    return smart_devices_cost


def calculate_surplus_controller_investment_cost(economic_parameters):
    """SURPLUS CONTROLLER."""
    if economic_parameters["surpluscontroller_bought"]:
        ccsc = json.load(open(file=path.join(hisim.utils.HISIMPATH['modular_household'], 'ComponentCostSurplusController.json'), mode='r'))
        surplus_controller_cost = ccsc["surplus_controller_cost"]
    else:
        surplus_controller_cost = 0
    return surplus_controller_cost


def calculate_heating_investment_cost(economic_parameters, heatpump_included, heater_capacity):
    """HEATING."""
    if economic_parameters["heatpump_bought"]:
        if not heatpump_included:
            hisim.log.information("Error: heatpump bought but not included")
        else:
            cchp = json.load(open(file=path.join(hisim.utils.HISIMPATH['modular_household'], 'ComponentCostHeatPump.json'), mode='r'))
            heatpump_cost_interp = scipy.interpolate.interp1d(cchp["capacity_for_cost"], cchp["cost_per_capacity"])
            heatpump_cost = heatpump_cost_interp(heater_capacity)
    else:
        heatpump_cost = 0
    return heatpump_cost


def calculate_battery_investment_cost(economic_parameters, battery_included, battery_capacity):
    """BATTERY."""
    if economic_parameters["battery_bought"]:
        if not battery_included:
            hisim.log.information("Error: battery bought but not included")
        else:
            ccb = json.load(open(file=path.join(hisim.utils.HISIMPATH['modular_household'], 'ComponentCostBattery.json'), mode='r'))
            battery_cost_interp = scipy.interpolate.interp1d(ccb["capacity_cost"], ccb["cost"])
            battery_cost = battery_cost_interp(battery_capacity)
    else:
        battery_cost = 0
    return battery_cost


def calculate_chp_investment_cost(economic_parameters, chp_included, chp_power):
    """CHP + H2 STORAGE + ELECTROLYSIS."""
    if economic_parameters["chp_bought"]:
        if not chp_included:
            hisim.log.information("Error: chp bought but not included")
        else:
            ccchp = json.load(open(file=path.join(hisim.utils.HISIMPATH['modular_household'], 'ComponentCostCHP.json'), mode='r'))
            chp_cost_interp = scipy.interpolate.interp1d(ccchp["capacity_for_cost"], ccchp["cost_per_capacity"])
            chp_cost = chp_cost_interp(chp_power)
    else:
        chp_cost = 0
    return chp_cost


def calculate_electrolyzer_investment_cost(economic_parameters, electrolyzer_included, electrolyzer_power):
    """Electrolyzer."""
    if economic_parameters["electrolyzer_bought"]:
        if not electrolyzer_included:
            hisim.log.information("Error: electrolyzer bought but not included")
        else:
            ccel = json.load(open(file=path.join(hisim.utils.HISIMPATH['modular_household'], 'ComponentCostElectrolyzer.json'), mode='r'))
            electrolyzer_cost_interp = scipy.interpolate.interp1d(ccel["capacity_for_cost"], ccel["cost_per_capacity"])
            electrolyzer_cost = electrolyzer_cost_interp(electrolyzer_power)
    else:
        electrolyzer_cost = 0
    return electrolyzer_cost


def calculate_h2storage_investment_cost(economic_parameters, h2system_included, h2_storage_size):
    """H2System."""
    if economic_parameters["h2system_bought"]:
        if not h2system_included:
            hisim.log.information("Error: h2system bought but not included")
        else:
            cch2 = json.load(open(file=path.join(hisim.utils.HISIMPATH['modular_household'], 'ComponentCostH2Storage.json'), mode='r'))
            h2_storage_cost_interp = scipy.interpolate.interp1d(cch2["capacity_for_cost"], cch2["cost_per_capacity"])
            h2_storage_cost = h2_storage_cost_interp(h2_storage_size)
    else:
        h2_storage_cost = 0
    return h2_storage_cost


def calculate_electric_vehicle_investment_cost(economic_parameters, ev_included, ev_capacity):
    """ELECTRIC VEHICLE."""
    if economic_parameters["ev_bought"]:
        if not ev_included:
            hisim.log.information("Error: EV bought but not included")
        else:
            ccev = json.load(open(file=path.join(hisim.utils.HISIMPATH['modular_household'], 'ComponentCostElectricVehicle.json'), mode='r'))
            ev_cost_interp = scipy.interpolate.interp1d(ccev["capacity_for_cost"], ccev["cost_per_capacity"])
            ev_cost = ev_cost_interp(ev_capacity)
    else:
        ev_cost = 0
    return ev_cost


def calculate_buffer_investment_cost(economic_parameters, buffer_included, buffer_volume):
    """BUFFER."""
    if economic_parameters["buffer_bought"]:
        if not buffer_included:
            hisim.log.information("Error: Buffer bought but not included")
        else:
            ccbu = json.load(open(file=path.join(hisim.utils.HISIMPATH['modular_household'], 'ComponentCostBuffer.json'), mode='r'))
            buffer_cost_interp = scipy.interpolate.interp1d(ccbu["capacity_for_cost"], ccbu["cost_per_capacity"])
            buffer_cost = buffer_cost_interp(buffer_volume)
    else:
        buffer_cost = 0
    return buffer_cost


def total_investment_cost_threshold_exceedance_check(economic_parameters, pv_cost, smart_devices_cost,
                                                     battery_cost, surplus_controller_cost, heatpump_cost, buffer_cost,
                                                     chp_cost, h2_storage_cost, electrolyzer_cost, ev_cost):
    """CHECKS WHETHER THE TOTAL COST OF INVESTMENT IS ABOVE THE INVESTMENT BUDGET AVAILABLE."""
    total_threshold = (economic_parameters["pv_threshold"] +
                       economic_parameters["smart_devices_threshold"] +
                       economic_parameters["heatpump_threshold"] +
                       economic_parameters["battery_threshold"] +
                       economic_parameters["buffer_threshold"] +
                       economic_parameters["h2system_threshold"] +
                       economic_parameters["chp_threshold"] +
                       economic_parameters["electrolyzer_threshold"] +
                       economic_parameters["surpluscontroller_threshold"] +
                       economic_parameters["ev_threshold"])
    hisim.log.information("total threshold" + str(total_threshold))

    investment_cost = pv_cost + smart_devices_cost + heatpump_cost + battery_cost + buffer_cost\
        + chp_cost + h2_storage_cost + electrolyzer_cost + ev_cost + surplus_controller_cost

    if investment_cost > total_threshold:
        hisim.log.information("Error: Total investment cost exceeded the total threshold")
    else:
        hisim.log.information("Everything alright: Total investment cost is below total threshold")
    return investment_cost


def investment_cost_per_component_exceedance_check(economic_parameters, pv_cost, smart_devices_cost, battery_cost, surplus_controller_cost,
                                                   heatpump_cost, buffer_cost, chp_cost, h2_storage_cost,
                                                   electrolyzer_cost, ev_cost):  # noqa: MC0001
    """CHECKS FOR EACH COMPONENT WHETHER THE INVESTMENT COST IS ABOVE THE THRESHOLD BUDGET."""
    if pv_cost > economic_parameters["pv_threshold"]:
        hisim.log.information("Problem: PV investment cost threshold exceeded.")
    elif smart_devices_cost > economic_parameters["smart_devices_threshold"]:
        hisim.log.information("Problem: Smart devices investment cost threshold exceeded.")
    elif heatpump_cost > economic_parameters["heatpump_threshold"]:
        hisim.log.information("Problem: Heatpump investment cost threshold exceeded.")
    elif battery_cost > economic_parameters["battery_threshold"]:
        hisim.log.information("Problem: Battery investment cost threshold exceeded.")
    elif buffer_cost > economic_parameters["buffer_threshold"]:
        hisim.log.information("Problem: Buffer investment cost threshold exceeded.")
    elif chp_cost > economic_parameters["chp_threshold"]:
        hisim.log.information("Problem: CHP investment cost threshold exceeded.")
    elif h2_storage_cost > economic_parameters["h2storage_threshold"]:
        hisim.log.information("Problem: H2Storage investment cost threshold exceeded.")
    elif electrolyzer_cost > economic_parameters["electrolyzer_threshold"]:
        hisim.log.information("Problem: Electrolyzer investment cost threshold exceeded.")
    elif ev_cost > economic_parameters["ev_threshold"]:
        hisim.log.information("Problem: EV investment cost threshold exceeded.")
    elif surplus_controller_cost > economic_parameters["surplus_controller_threshold"]:
        hisim.log.information("Problem: Surplus controller investment cost threshold exceeded.")
    else:
        hisim.log.information("All components do not exceed their treshold.")



# Owned
from hisim.simulationparameters import SimulationParameters

from dataclasses import dataclass
from dataclasses_json import dataclass_json
from hisim.component import Component, SingleTimeStepValues, ComponentInput, ComponentOutput

from hisim import loadtypes as lt

from hisim.components.configuration import PhysicsConfig
from hisim import log
from typing import List, Any
__authors__ = "Frank Burkrad, Maximilian Hillen"
__copyright__ = "Copyright 2021, the House Infrastructure Project"
__credits__ = ["Noah Pflugradt"]
__license__ = ""
__version__ = ""
__maintainer__ = "Maximilian Hillen"
__email__ = "maximilian.hillen@rwth-aachen.de"
__status__ = ""
#

@dataclass_json
@dataclass
class ElectrolyzerWithStorageConfig:
    component_name: str
    waste_energy: float # [W]
    min_power: float # [W]
    max_power: float # [W]
    min_power_percent: float # [%]
    max_power_percent: float # [W]
    min_hydrogen_production_rate_hour: float # [Nl/h]
    max_hydrogen_production_rate_hour: float # [Nl/h]
    pressure_hydrogen_output: float # [bar]

@dataclass_json
@dataclass
class ElectrolyzerWithHydrogenStorageConfig:
    component_name: str
    min_capacity: float # [kg_H2]
    max_capacity: float # [kg_H2]
    starting_fill: float # [kg_H2]
    max_charging_rate_hour: float # [kg/h]
    max_discharging_rate_hour: float # [kg/h]
    energy_for_charge: float # [kWh/kg]
    energy_for_discharge: float # [kWh/kg]
    loss_factor_per_day: float # [lost_%/day]

class ElectrolyzerSimulation:
    def __init__(self,
                 waste_energy : float,
                 min_power: float,
                 max_power: float,
                 min_power_percent: float,
                 max_power_percent: float,
                 min_hydrogen_production_rate_hour: float,
                 max_hydrogen_production_rate_hour: float,
                 min_hydrogen_production_rate: float,
                 max_hydrogen_production_rate: float,
                 pressure_hydrogen_output: float):
        """
        The electrolyzer converts electrical energy [kWh] into hydrogen [kg]
        It can work in a certain range from x to 100% or be switched off = 0%
        The conversion rate is given by the supplier and is directly used
            maybe a change to efficiency can be made but its just making things more complex with no benefit
        Between the given values, the values are calculated by an interpolation.
            --> If the load curve is linear a fixed factor could be calculated.

        Therefore it has an operational state
        All the min values and  all the max values are connected and the electrolyzer can operate between them.

        The waste energy in electolyzers is not used to provide heat for the households demand
        Output pressure may be used in the future for the
        """
        self.state = 0
        self.min_power = min_power
        self.max_power = max_power
        self.min_power_percent = min_power_percent
        self.max_power_percent = max_power_percent

        self.min_hydrogen_production_rate = min_hydrogen_production_rate
        self.max_hydrogen_production_rate = max_hydrogen_production_rate
        self.min_hydrogen_production_rate_hour = min_hydrogen_production_rate
        self.max_hydrogen_production_rate_hour = max_hydrogen_production_rate
        self.waste_energy = waste_energy
        self.pressure_hydrogen_output = pressure_hydrogen_output   # not used so far

    def convert_electricity(self, electricity_input, seconds_per_timestep, hydrogen_not_stored ):
        """
        Electricity from electricity distributor (combination of PV, Grid, CHP and demand) will be converted to hydrogen

        Check if electricity input is inside the power range of electrolyzer and eventually subtract the unusable power.
        Calculate the current power level of the electrolyzer in this timestep
        Calculate the current hydrogen output in liter
        Conversion to kg_H2
        calculate the Oxygen output by mass fraction in the water molecule

        :param      electricity_input:      Electricity from electricity distributor [W]
        :return:    hydrogen_output [kg]    Amount of produced hydrogen
                    oxygen_output [kg]      Amount of produced oxygen
                    unused_power [W]        Unused power if electricity_input is outside the operation range
        """

        # interpolation between two points. Future: add input data with higher resolution.. but not provided
        # interpolation for current power level & hydrogen output
        power_level = self.min_power_percent + (100 - self.min_power_percent) * (electricity_input- self.min_power) / (self.max_power - self.min_power)
        assert self.min_power_percent <= power_level <= self.max_power_percent
        # Nl / s
        hydrogen_output_liter = self.min_hydrogen_production_rate + ((self.max_hydrogen_production_rate - self.min_hydrogen_production_rate) * (power_level - self.min_power_percent) / (self.max_power_percent - self.min_power_percent))
        assert self.min_hydrogen_production_rate <= hydrogen_output_liter <= self.max_hydrogen_production_rate
        # kg/s = l/s / 1000 * kg/m³
        hydrogen_output = (hydrogen_output_liter / 1000) * PhysicsConfig.hydrogen_density
        oxygen_output = hydrogen_output * (88.8 / 11.2)

        if hydrogen_not_stored>0:
            hydrogen_real= hydrogen_output - hydrogen_not_stored
            if hydrogen_real==0:
                return hydrogen_output, 0, 0, 0
            hydrogen_output_liter_real = hydrogen_real * 1000 / PhysicsConfig.hydrogen_density
            power_level_real = self.min_power_percent + (
                        hydrogen_output_liter_real - self.min_hydrogen_production_rate) * (
                                           self.max_power_percent - self.min_power_percent) / (
                                           self.max_hydrogen_production_rate - self.min_hydrogen_production_rate)
            electricity_input_real = self.min_power + (self.max_power - self.min_power) * (
                        power_level_real - self.min_power_percent) / (100 - self.min_power_percent)
            oxygen_output = hydrogen_real * (88.8 / 11.2)

            return hydrogen_output, oxygen_output, power_level_real,electricity_input_real
        elif hydrogen_not_stored<0:
            log.error("Error")
            raise ValueError("hydrogen was leftover")
        elif hydrogen_not_stored == 0:
            electricity_input_real=electricity_input
            return hydrogen_output, oxygen_output, power_level,electricity_input_real


        """
        # if there are more datapoints given
        # power_datapoints [60%, 70%,...,100]
        # output_datapoints [300Nl/h, 350Nl/h, 500Nl/h]

        if power_level in power_datapoints:
            for i in power_datapoints:
                if power_level == power_datapoints[i]:
                    hydrogen_output = output_datapiont[i]
        else:
            # get closest values
            for i in power_datapoints:
                if power_level > power_datapoints[i]:
                    lower_datapoint = i
                if power_level < power_datapoints[i]:
                    upper_datapoint = i
                    break
            # interpolation
            hydrogen_output = output_datapoints[lower_datapoint] + (
                    output_datapoints[upper_datapoint] - output_datapoints[lower_datapoint]) * (
                                      power_datapoints[upper_datapoint] - power_level) / (
                                      power_datapoints[upper_datapoint] - (power_datapoints[lower_datapoint])
        """

        # Water consists out of H2O. The molecular weight is 11.2% Hydrogen and 88.8% Oxygen.
        # This means: From the calculated hydrogen mass, the oxygen mass can be inferred
        # this is just a fun fact and is not needed for simulation
        oxygen_output = hydrogen_output * (88.8 / 11.2)

        return hydrogen_output, oxygen_output, power_level

class AdvancedElectrolyzer(Component):
    # input
    ElectricityInput = "Electricity Input"              # W
    HydrogenNotStored= "HydrogenNotStored"            # kg/s
    # output
    WaterDemand = "Water Demand"                        # kg/s
    HydrogenOutput = "Hydrogen Output"                  # kg/s
    OxygenOutput = "Oxygen Output"                      # kg/s
    EnergyLosses = "Energy Losses"                      # W
    UnusedPower = "Unused Power"                        # W
    ElectrolyzerEfficiency = "Electrolyzer Efficiency"  # -
    PowerLevel = "Power Level"                          # %
    ElectricityRealNeeded = "ElectricityRealNeeded"

    def __init__(self, my_simulation_parameters: SimulationParameters,
                 config: ElectrolyzerWithStorageConfig):

        super().__init__(config.component_name, my_simulation_parameters=my_simulation_parameters)

        # input
        self.hydrogen_not_stored: ComponentInput = self.add_input(self.component_name, AdvancedElectrolyzer.HydrogenNotStored, lt.LoadTypes.HYDROGEN, lt.Units.KG, True)

        self.electricity_input: ComponentInput = self.add_input(self.component_name, AdvancedElectrolyzer.ElectricityInput, lt.LoadTypes.ELECTRICITY, lt.Units.WATT, True)
        # output
        self.water_demand: ComponentOutput = self.add_output(self.component_name, AdvancedElectrolyzer.WaterDemand, lt.LoadTypes.WATER, lt.Units.KG_PER_SEC)
        self.hydrogen_output: ComponentOutput = self.add_output(self.component_name, AdvancedElectrolyzer.HydrogenOutput, lt.LoadTypes.HYDROGEN, lt.Units.KG_PER_SEC)
        self.oxygen_output: ComponentOutput = self.add_output(self.component_name, AdvancedElectrolyzer.OxygenOutput, lt.LoadTypes.OXYGEN, lt.Units.KG_PER_SEC)
        self.energy_losses: ComponentOutput = self.add_output(self.component_name, AdvancedElectrolyzer.EnergyLosses, lt.LoadTypes.ELECTRICITY, lt.Units.WATT)
        self.unused_power: ComponentOutput = self.add_output(self.component_name, AdvancedElectrolyzer.UnusedPower, lt.LoadTypes.ELECTRICITY, lt.Units.WATT)
        self.electricity_real_needed: ComponentOutput = self.add_output(self.component_name, AdvancedElectrolyzer.ElectricityRealNeeded, lt.LoadTypes.ELECTRICITY, lt.Units.WATT)

        self.electrolyzer_efficiency: ComponentOutput = self.add_output(self.component_name, AdvancedElectrolyzer.ElectrolyzerEfficiency, lt.LoadTypes.ANY, lt.Units.ANY)
        self.power_level: ComponentOutput = self.add_output(self.component_name, AdvancedElectrolyzer.PowerLevel, lt.LoadTypes.ANY, lt.Units.PERCENT)

        self.max_power:float = config.max_power
        self.min_power = config.min_power
        self.waste_energy = config.waste_energy
        min_hydrogen_production_rate = config.min_hydrogen_production_rate_hour / 3600  # [Nl/s]
        max_hydrogen_production_rate = config.max_hydrogen_production_rate_hour / 3600  # [Nl/s]

        self.electrolyzer = ElectrolyzerSimulation(waste_energy = config.waste_energy,
                                                   min_power =config.min_power,
                                                   max_power =config.max_power,
                                                   min_power_percent =config.min_power_percent,
                                                   max_power_percent =config.max_power_percent,
                                                   min_hydrogen_production_rate_hour =config.min_hydrogen_production_rate_hour,
                                                   max_hydrogen_production_rate_hour =config.max_hydrogen_production_rate_hour,
                                                   min_hydrogen_production_rate =min_hydrogen_production_rate,
                                                   max_hydrogen_production_rate =max_hydrogen_production_rate,
                                                   pressure_hydrogen_output =config.pressure_hydrogen_output)
        self.seconds_per_timestep = my_simulation_parameters.seconds_per_timestep
        self.previous_state = 0
    def write_to_report(self):
        lines = []
        lines.append("CHP operation with constant electical and thermal power: " + self.component_name)
        return lines
    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass
    @staticmethod
    def get_default_config():
        config=ElectrolyzerWithStorageConfig(
            component_name= "Electrolyzer",
            waste_energy=400,  # [W]
            min_power= 1_200,  # [W]
            max_power= 2_400,  # [W]
            min_power_percent = 60,  # [%],
            max_power_percent = 100,  # [W]
            min_hydrogen_production_rate_hour = 300,  # [Nl/h]
            max_hydrogen_production_rate_hour = 5000,  # [Nl/h]
            pressure_hydrogen_output = 30  # [bar]
            )
        return config
    def i_save_state(self):
        self.previous_state = self.electrolyzer.state

    def i_restore_state(self):
        self.electrolyzer.state = self.previous_state

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool)->None:
        electricity_input:float = stsv.get_input_value( self.electricity_input )
        hydrogen_input = stsv.get_input_value( self.hydrogen_not_stored )

        if electricity_input < 0:
            raise Exception( "trying to run electrolyzer with negative amount" + str( electricity_input ) )

        # operations can be skipped if there is no action (electricity_input == 0) in this timestep
        # maybe advance this to electricity_input >= ElectrolyzerConfig.min_power
        hydrogen_output = 0
        oxygen_output = 0
        losses_this_timestep:float = 0
        unused_power: float = 0
        power_level = 0
        # the following is already regulated in the electricity distributor

        if electricity_input > self.max_power:
            unused_power = electricity_input - self.max_power
            electricity_input = self.max_power
        elif 0 <= electricity_input < self.min_power:
            unused_power = electricity_input
            electricity_input = 0
        if electricity_input >= self.min_power:
            hydrogen_output, oxygen_output, power_level, electricity_needed = \
                self.electrolyzer.convert_electricity( electricity_input, self.seconds_per_timestep, hydrogen_input / self.seconds_per_timestep )
            # the losses ae included in the efficiency providedd by supplyer and are not calculated separately
            losses_this_timestep = self.waste_energy
            # unused_hydrogen = charging_amount - hydrogen_input  # add if needed?
            unused_power = unused_power + ( electricity_input - electricity_needed )
        electricity_real_needed = stsv.get_input_value( self.electricity_input ) - unused_power
        # water is split into these products
        if oxygen_output==0:
            water_consumption=0
        else:
            water_consumption = hydrogen_output + oxygen_output
        try:
            # -/- = kg/s * J/kg / W
            electrolyzer_efficiency = (hydrogen_output * PhysicsConfig.hydrogen_specific_fuel_value_per_kg) / electricity_input
            assert self.min_power <= electricity_input <= self.max_power
        except ZeroDivisionError:
            electrolyzer_efficiency = 0
            assert electricity_input < self.min_power

        stsv.set_output_value( self.water_demand, water_consumption )
        stsv.set_output_value( self.hydrogen_output, hydrogen_output )
        stsv.set_output_value( self.oxygen_output, oxygen_output )
        stsv.set_output_value( self.energy_losses, losses_this_timestep )

        stsv.set_output_value( self.electricity_real_needed, electricity_real_needed )
        stsv.set_output_value( self.unused_power, unused_power )
        stsv.set_output_value( self.electrolyzer_efficiency, electrolyzer_efficiency )
        stsv.set_output_value( self.power_level, power_level )

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues)->None:
        # alle ausgabewerte die zu überprüfen sind können hiermit fehlerausgabeüberprüft werden
        pass

class HydrogenStorageSimulation:
    """
    Hydrogen storage to store hydrogen which is produced by the electrolyzer.
    The charging and sometimes discharging of the tank is connected with an electricity (energy) consumption. (Compressor, cooling,...). This energy is given by a fixed value which is not related to the actual fill level.
    Losses are related to the fill level and are given in lost%/24h.
    """
    def __init__(self,
                 fill:float,
                 max_capacity: float,
                 min_capacity: float,
                 max_charging_rate: float,
                 max_discharging_rate: float,
                 energy_to_charge: float,
                 energy_to_discharge: float,
                 loss_factor: float
                 ):
        """
        The storage has a minimum and maximum storage level and a current fill status
        The fill status must be between the min and max capacity
        A maximum charging and discharging rate restrict these processes.
        """
        self.fill=fill
        self.max_capacity=max_capacity
        self.min_capacity=min_capacity
        self.max_charging_rate=max_charging_rate
        self.max_discharging_rate=max_discharging_rate
        self.energy_to_charge=energy_to_charge
        self.energy_to_discharge=energy_to_discharge
        self.loss_factor=loss_factor

        #assert self.max_charging_rate > (ElectrolyzerConfig.max_hydrogen_production_rate / 1000 * PhysicsConfig.hydrogen_density)
        #assert self.max_discharging_rate > chp.CHPConfig.P_total_max / PhysicsConfig.hydrogen_specific_fuel_value_per_kg

    def store(self, hydrogen_input: float, seconds_per_timestep: int)->Any:
        """
        Notice: Write return statement and the function goes back to to the caller method immediately
        Storage function:
        The maximum possible charging amount is calculated.
        The hydrogen inflow is restricted by this. If its too high the delta cant be stored.

        Calculate the new fill level.
        Possible restrictions due to a full tank.
            Nothing can be loaded
            Parts of the hydrogen inflow can be loaded.
        Calculate the energy consumption which is needed for the process.

        :param      hydrogen_input:         Hydrogen to the tank [kg_H2]
        :param      seconds_per_timestep:   Seconds in this timestep [s]
        :return:    hydrogen_input [kg]     Amount of hydrogen which was stored in this timestep
                    energy_demand [W]       Consumed Energy for storing
                    delta_not_stored [kg]   Amount of hydrogen which was not stored in this timestep
        """
        max_charging = self.max_charging_rate * seconds_per_timestep

        delta_not_stored:float = 0
        if hydrogen_input > max_charging:
            delta_not_stored = hydrogen_input - max_charging
            hydrogen_input = max_charging

        if self.fill + hydrogen_input < self.max_capacity:
            # fits completely
            self.fill += hydrogen_input
            # W = kg * kWh/kg * 3600_s/h * 1000_1/k / s
            energy_demand = hydrogen_input * self.energy_to_charge * 3600 * 1000 / seconds_per_timestep
            return hydrogen_input, energy_demand, delta_not_stored
        elif self.fill >= self.max_capacity:
            # tank is already full
            hydrogen_input = 0
            energy_demand = 0
            # limitation by tanksize and charging rate
            delta_not_stored += hydrogen_input
            return hydrogen_input, energy_demand, delta_not_stored
        if self.fill < self.max_capacity:
            # fits partially
            # returns amount which an be put in
            amount_stored = self.max_capacity - self.fill
            self.fill += amount_stored
            delta_not_stored = hydrogen_input - amount_stored
            energy_demand = amount_stored * self.energy_to_charge * 3600 * 1000 / seconds_per_timestep    # w
            return amount_stored, energy_demand, delta_not_stored
        raise Exception("forgotten case")

    def withdraw(self, hydrogen_output: float, seconds_per_timestep: int) -> Any:
        """
        Discharging function. Functionality its the reverse of the store function.
        :param      hydrogen_output:        Demand on hydrogen [kg_H2]
        :param      seconds_per_timestep:   Seconds in this timestep [s]
        :return:    hydrogen_output [kg]    Amount of hydrogen which was released in this timestep
                    energy_demand [W]       Consumed Energy for release
                    delta_not_released [kg] Amount of hydrogen which couldn't be released in this timestep
        """
        max_discharging = self.max_discharging_rate * seconds_per_timestep
        delta_not_released:float = 0
        if hydrogen_output > max_discharging:
            delta_not_released = hydrogen_output - max_discharging
            hydrogen_output = max_discharging

        if self.fill > hydrogen_output:
            # has enough
            self.fill -= hydrogen_output
            energy_demand = hydrogen_output * self.energy_to_discharge * 3600 * 1000 / seconds_per_timestep   # W
            return hydrogen_output, energy_demand, delta_not_released
        if self.fill <= self.min_capacity:
            # empty
            energy_demand = 0
            delta_not_released = hydrogen_output
            return 0, energy_demand, delta_not_released
        if self.fill < hydrogen_output:
            # can provide hydrogen partially,
            # added recently :but in this case to simplify work of CHP, say that no hydrogen can be provided
            amount = self.fill
            delta_not_released = hydrogen_output - self.fill
            self.fill = 0
            energy_demand = amount * self.energy_to_discharge * 3600 * 1000 / seconds_per_timestep    # W
            return 0, 0, delta_not_released
        raise Exception("forgotten case")

    def storage_losses(self, seconds_per_timestep):
        """
        How much hydrogen is lost in the time period:
        Liquid tanks:  evaporation
        Pressure tanks: leakages
        metalhydride:   suppliers say there are no losses (??)

        The losses per day are given in %. This means a full tank has more losses compared to a half full tank.
        First, the given loss factor is converted into the losses in the timestep.
        Loses this timestep are calculated and subtracted from the hydrogen storage
        :param      seconds_per_timestep:   Seconds in this timestep [s]
        :return:    losses in this timestep in kg
        """
        # conversion from lost_%/24h to lost_%/timestep

        hydrogen_losses_this_timestep = self.loss_factor / 24 * (seconds_per_timestep / 3600)
        # % have to be divided by 100
        losses_this_timestep = self.fill * (hydrogen_losses_this_timestep / 100)
        self.fill -= losses_this_timestep
        return losses_this_timestep

class HydrogenStorage(Component):
    # input
    ChargingHydrogenAmount = "Charging Hydrogen Amount"                     # kg/s
    DischargingHydrogenAmountTarget = "DischargingHydrogenAmountTarget"               # kg/s
    # output
    CurrentHydrogenFillLevel = "Current Hydrogen Fill Level Absolute"       # kg
    CurrentHydrogenFillLevelPercent = "Current Hydrogen Fill Level Percent" # %
    StorageDelta = "Storage Delta"                                          # kg
    HydrogenNotStored = "Hydrogen Not Stored"                               # kg
    HydrogenNotReleased = "Hydrogen Not Released"                           # kg

    HydrogenStorageEnergyDemand = "Hydrogen Storage Energy Demand"          # W
    HydrogenLosses = "Hydrogen Losses"                                      # kg
    DischargingHydrogenAmountReal= "Discharging Hydrogen Amount Real"               # kg/s

    def __init__(self, my_simulation_parameters:SimulationParameters,
                 config:ElectrolyzerWithHydrogenStorageConfig):
        super().__init__(config.component_name, my_simulation_parameters=my_simulation_parameters)
        self.charging_hydrogen: ComponentInput = self.add_input(self.component_name, HydrogenStorage.ChargingHydrogenAmount, lt.LoadTypes.HYDROGEN, lt.Units.KG_PER_SEC, True)
        self.discharging_hydrogen: ComponentInput = self.add_input(self.component_name, HydrogenStorage.DischargingHydrogenAmountTarget, lt.LoadTypes.HYDROGEN, lt.Units.KG_PER_SEC, False)

        self.current_fill: ComponentOutput = self.add_output(self.component_name, HydrogenStorage.CurrentHydrogenFillLevel, lt.LoadTypes.HYDROGEN, lt.Units.KG)
        self.current_fill_percent: ComponentOutput = self.add_output(self.component_name, HydrogenStorage.CurrentHydrogenFillLevelPercent, lt.LoadTypes.HYDROGEN, lt.Units.PERCENT)
        self.storage_delta: ComponentOutput = self.add_output(self.component_name, HydrogenStorage.StorageDelta, lt.LoadTypes.HYDROGEN, lt.Units.KG_PER_SEC)
        self.hydrogen_not_stored: ComponentOutput = self.add_output(self.component_name, HydrogenStorage.HydrogenNotStored, lt.LoadTypes.HYDROGEN, lt.Units.KG)
        self.hydrogen_not_released: ComponentOutput = self.add_output(self.component_name, HydrogenStorage.HydrogenNotReleased, lt.LoadTypes.HYDROGEN, lt.Units.KG)
        self.hydrogen_storage_energy_demand: ComponentOutput = self.add_output(self.component_name, HydrogenStorage.HydrogenStorageEnergyDemand, lt.LoadTypes.ELECTRICITY, lt.Units.WATT)
        self.hydrogen_losses: ComponentOutput = self.add_output(self.component_name, HydrogenStorage.HydrogenLosses, lt.LoadTypes.HYDROGEN, lt.Units.KG)
        self.discharging_hydrogen_real: ComponentOutput = self.add_output(
            object_name=self.component_name, field_name=HydrogenStorage.DischargingHydrogenAmountReal, load_type=lt.LoadTypes.HYDROGEN, 
            unit=lt.Units.KG_PER_SEC, sankey_flow_direction=False)

        self.max_capacity=config.max_capacity
        self.seconds_per_timestep = my_simulation_parameters.seconds_per_timestep
        self.previous_state:float = 0
        self.max_charging_rate = config.max_charging_rate_hour / 3600
        self.max_discharging_rate = config.max_discharging_rate_hour / 3600

        self.hydrogenstorage = HydrogenStorageSimulation(fill=config.starting_fill,
                                                         max_capacity=config.max_capacity ,
                                                         min_capacity=config.min_capacity,
                                                         max_charging_rate=self.max_charging_rate,
                                                         max_discharging_rate=self.max_discharging_rate,
                                                         energy_to_charge=config.energy_for_charge,
                                                         energy_to_discharge=config.energy_for_discharge,
                                                         loss_factor=config.loss_factor_per_day
                                                         )

    @staticmethod
    def get_default_config():
        config=ElectrolyzerWithHydrogenStorageConfig(
            component_name = "HydrogenStorage",
            min_capacity = 0,
            max_capacity = 500,
            starting_fill = 400,
            max_charging_rate_hour = 2,
            max_discharging_rate_hour = 2,
            energy_for_charge = 0,
            energy_for_discharge= 0,
            loss_factor_per_day= 0)
        return config
    def i_save_state(self):
        self.previous_state = self.hydrogenstorage.fill
    def i_prepare_simulation(self) -> None:
        """ Prepares the simulation. """
        pass
    def i_restore_state(self):
        self.hydrogenstorage.fill = self.previous_state

    def i_simulate(self, timestep: int, stsv: SingleTimeStepValues, force_convergence: bool) -> None:
        # Setting up the internal values
        charging_amount_sec = stsv.get_input_value(self.charging_hydrogen)
        charging_amount = charging_amount_sec * self.seconds_per_timestep
        # takes positive value for the discharge
        discharging_amount_sec = stsv.get_input_value(self.discharging_hydrogen)
        discharging_amount = discharging_amount_sec * self.seconds_per_timestep

        if charging_amount < 0:
            raise Exception("trying to charge with negative amount" + str(charging_amount))
        if discharging_amount < 0:
            raise Exception("trying to discharge with negative amount: " + str(discharging_amount))
        # operations can be skipped if there is no action in this timestep
        hydrogen_input = 0
        hydrogen_output = 0
        charging_energy_demand = 0
        discharging_energy_demand = 0
        hydrogen_not_stored:float = 0
        hydrogen_not_released = 0

        if charging_amount > 0 and discharging_amount > 0:
            # simultaneous charging and discharging has to be prevented
            # hydrogen can be used directly
            delta = charging_amount - discharging_amount
            if delta >= 0:
                charging_amount = delta
                discharging_amount = 0
            else:
                charging_amount = 0
                discharging_amount = -delta
        if charging_amount > 0 and discharging_amount > 0:
            # to check previous if command
            # a heat stick in the storage tank would be better --> direct use of el.Energy
            raise Exception("Tank cant be charged and discharged in the same timestep. Use existing Hydrogen! Delta:" + str(delta))

        if charging_amount > 0:
            hydrogen_input, charging_energy_demand, hydrogen_not_stored = self.hydrogenstorage.store(charging_amount, self.seconds_per_timestep)
            # unused_hydrogen = charging_amount - hydrogen_input  # add if needed?
        if discharging_amount > 0:
            hydrogen_output, discharging_energy_demand, hydrogen_not_released = self.hydrogenstorage.withdraw(discharging_amount, self.seconds_per_timestep)
            if hydrogen_not_released>0:
                hydrogen_output=0
                discharging_energy_demand=0
        losses_this_timestep = self.hydrogenstorage.storage_losses(self.seconds_per_timestep)
        # discharging amount target
        energy_demand = charging_energy_demand + discharging_energy_demand          # W
        actual_delta = (hydrogen_input - hydrogen_output - losses_this_timestep) / self.seconds_per_timestep      # kg/s
        percent_fill = self.hydrogenstorage.fill / self.max_capacity

        hydrogen_not_stored = stsv.get_input_value(self.charging_hydrogen) * self.seconds_per_timestep-hydrogen_input

        stsv.set_output_value(self.hydrogen_storage_energy_demand, energy_demand)
        stsv.set_output_value(self.hydrogen_losses, losses_this_timestep)
        stsv.set_output_value(self.storage_delta, actual_delta)
        stsv.set_output_value(self.hydrogen_not_stored, hydrogen_not_stored)
        stsv.set_output_value(self.hydrogen_not_released, hydrogen_not_released)
        stsv.set_output_value(self.current_fill, self.hydrogenstorage.fill)
        stsv.set_output_value(self.current_fill_percent, percent_fill)

    def i_doublecheck(self, timestep: int, stsv: SingleTimeStepValues) -> None:
        # alle ausgabewerte die zu überprüfen sind können hiermit fehlerausgabeüberprüft werden
        pass

    def write_to_report(self) -> List[str]:
        lines = []
        lines.append("Hydrogen Storage: " + self.component_name)
        return lines





""" Contains a component that uses the UTSP to provide LoadProfileGenerator data. """
import datetime
import errno
import io
import itertools
import json
import os
from dataclasses import dataclass
from typing import Dict, Optional, Tuple

import numpy as np
import pandas as pd
from dataclasses_json import dataclass_json

from utspclient import client, datastructures, result_file_filters
from utspclient.helpers import lpg_helper
from utspclient.helpers.lpgdata import (
    ChargingStationSets,
    Households,
    HouseTypes,
    LoadTypes,
    TransportationDeviceSets,
    TravelRouteSets,
)
from utspclient.helpers.lpgpythonbindings import CalcOption, JsonReference

# Owned
from hisim import component as cp
from hisim import loadtypes as lt
from hisim import log, utils
from hisim.components.configuration import HouseholdWarmWaterDemandConfig, PhysicsConfig
from hisim.simulationparameters import SimulationParameters


@dataclass_json
@dataclass
class UtspLpgConnectorConfig:

    """Config class for UtspLpgConnector. Contains LPG parameters and UTSP connection parameters."""

    url: str
    api_key: str
    household: JsonReference
    result_path: str
    travel_route_set: JsonReference
    transportation_device_set: JsonReference
    charging_station_set: JsonReference

    @staticmethod
    def get_default_config(
        url: str = "http://localhost:443/api/v1/profilerequest", api_key: str = ""
    ) -> "UtspLpgConnectorConfig":
        """Creates a default configuration. Chooses default values for the LPG parameters."""
        result_path = os.path.join(utils.get_input_directory(), "lpg_profiles")
        config = UtspLpgConnectorConfig(
            url,
            api_key,
            Households.CHR01_Couple_both_at_Work,
            result_path,
            travel_route_set=TravelRouteSets.Travel_Route_Set_for_10km_Commuting_Distance,
            transportation_device_set=TransportationDeviceSets.Bus_and_one_30_km_h_Car,
            charging_station_set=ChargingStationSets.Charging_At_Home_with_03_7_kW,
        )
        return config


class UtspLpgConnector(cp.Component):

    """Component that provides data from the LoadProfileGenerator.

    This component provides the heating generated, the electricity and water consumed
    by the residents. Furthermore, transportation and device flexibility data is stored
    in separate files, configurable via the config object.
    The data is retrieved from the UTSP, which executes the LoadProfileGenerator to simulate
    the specified household.
    """

    # Inputs
    WW_MassInput = "Warm Water Mass Input"  # kg/s
    WW_TemperatureInput = "Warm Water Temperature Input"  # °C

    # Outputs
    # output
    WW_MassOutput = "Mass Output"  # kg/s
    WW_TemperatureOutput = "Temperature Output"  # °C
    EnergyDischarged = "Energy Discharged"  # W
    DemandSatisfied = "Demand Satisfied"  # 0 or 1

    NumberByResidents = "NumberByResidents"
    HeatingByResidents = "HeatingByResidents"
    ElectricityOutput = "ElectricityOutput"
    WaterConsumption = "WaterConsumption"

    Electricity_Demand_Forecast_24h = "Electricity_Demand_Forecast_24h"

    # Similar components to connect to:
    # None
    @utils.measure_execution_time
    def __init__(
        self,
        my_simulation_parameters: SimulationParameters,
        config: UtspLpgConnectorConfig,
    ) -> None:
        """Initializes the component and retrieves the LPG data."""
        super().__init__(
            name=UtspLpgConnector.__name__,
            my_simulation_parameters=my_simulation_parameters,
        )
        self.utsp_config = config
        self.build()

        # Inputs - Not Mandatory
        self.ww_mass_input: cp.ComponentInput = self.add_input(
            self.component_name,
            self.WW_MassInput,
            lt.LoadTypes.WARM_WATER,
            lt.Units.KG_PER_SEC,
            False,
        )
        self.ww_temperature_input: cp.ComponentInput = self.add_input(
            self.component_name,
            self.WW_TemperatureInput,
            lt.LoadTypes.WARM_WATER,
            lt.Units.CELSIUS,
            False,
        )

        self.number_of_residents_c: cp.ComponentOutput = self.add_output(
            self.component_name, self.NumberByResidents, lt.LoadTypes.ANY, lt.Units.ANY
        )
        self.heating_by_residents_c: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.HeatingByResidents,
            lt.LoadTypes.HEATING,
            lt.Units.WATT,
        )
        self.electricity_output_c: cp.ComponentOutput = self.add_output(
            object_name=self.component_name,
            field_name=self.ElectricityOutput,
            load_type=lt.LoadTypes.ELECTRICITY,
            unit=lt.Units.WATT,
            postprocessing_flag=[
                lt.InandOutputType.ELECTRICITY_CONSUMPTION_UNCONTROLLED
            ],
        )

        self.water_consumption_c: cp.ComponentOutput = self.add_output(
            self.component_name,
            self.WaterConsumption,
            lt.LoadTypes.WARM_WATER,
            lt.Units.LITER,
        )

    def i_save_state(self) -> None:
        """Empty method as component has no state."""
        pass

    def i_restore_state(self) -> None:
        """Empty method as component has no state."""
        pass

    def i_prepare_simulation(self) -> None:
        """Prepares the simulation."""
        pass

    def i_doublecheck(self, timestep: int, stsv: cp.SingleTimeStepValues) -> None:
        """Gets called after the iterations are finished at each time step for potential debugging purposes."""
        pass

    def i_simulate(self, timestep: int, stsv: cp.SingleTimeStepValues, force_convergence: bool) -> None:
        """Sets the current output values with data retrieved during initialization."""
        if self.ww_mass_input.source_output is not None:
            # ww demand
            ww_temperature_demand = HouseholdWarmWaterDemandConfig.ww_temperature_demand

            # From Thermal Energy Storage
            ww_mass_input_per_sec = stsv.get_input_value(self.ww_mass_input)  # kg/s
            # ww_mass_input = ww_mass_input_per_sec * self.seconds_per_timestep           # kg
            ww_mass_input: float = ww_mass_input_per_sec
            ww_temperature_input = stsv.get_input_value(self.ww_temperature_input)  # °C

            # Information import
            freshwater_temperature = (
                HouseholdWarmWaterDemandConfig.freshwater_temperature
            )
            temperature_difference_cold = (
                HouseholdWarmWaterDemandConfig.temperature_difference_cold
            )
            energy_losses = 0
            specific_heat = 4180 / 3600

            ww_energy_demand = (
                specific_heat
                * self.water_consumption[timestep]
                * (ww_temperature_demand - freshwater_temperature)
            )

            if ww_energy_demand > 0 and (
                ww_mass_input == 0 and ww_temperature_input == 0
            ):
                """first iteration --> random numbers"""
                ww_temperature_input = 40.45
                ww_mass_input = 9.3

            """
            Warm water is provided by the warmwater stoage.
            The household needs water at a certain temperature. To get the correct temperature the amount of water from
            the wws is regulated and is depending on the temperature provided by the wws. The backflowing water to wws
            is cooled down to the temperature of (freshwater+temperature_difference_cold) --> ww_temperature_output.
            """
            if ww_energy_demand > 0:
                # heating up the freshwater. The mass is consistent
                energy_discharged = ww_energy_demand + energy_losses
                ww_temperature_output: float = (
                    freshwater_temperature + temperature_difference_cold
                )
                ww_mass_input = energy_discharged / (
                    PhysicsConfig.water_specific_heat_capacity_in_joule_per_kilogram_per_kelvin
                    * (ww_temperature_input - ww_temperature_output)
                )
            else:
                ww_temperature_output = ww_temperature_input
                ww_mass_input = 0
                energy_discharged = 0

        stsv.set_output_value(
            self.number_of_residents_c, self.number_of_residents[timestep]
        )
        stsv.set_output_value(
            self.heating_by_residents_c, self.heating_by_residents[timestep]
        )
        stsv.set_output_value(
            self.electricity_output_c, self.electricity_consumption[timestep]
        )
        stsv.set_output_value(
            self.water_consumption_c, self.water_consumption[timestep]
        )

        if self.my_simulation_parameters.system_config.predictive:
            last_forecast_timestep = int(
                timestep
                + 24 * 3600 / self.my_simulation_parameters.seconds_per_timestep
            )
            if last_forecast_timestep > len(self.electricity_consumption):
                last_forecast_timestep = len(self.electricity_consumption)
            demandforecast = self.electricity_consumption[
                timestep:last_forecast_timestep
            ]
            self.simulation_repository.set_entry(
                self.Electricity_Demand_Forecast_24h, demandforecast
            )

    def get_resolution(self) -> str:
        """Gets the temporal resolution of the simulation as a string in the format hh:mm:ss.

        :return: resolution of the simulation
        :rtype: str
        """
        seconds = self.my_simulation_parameters.seconds_per_timestep
        resolution = datetime.timedelta(seconds=seconds)
        return str(resolution)

    def get_profiles_from_utsp(self) -> Tuple[str, str, str, str]:
        """Requests the required load profiles from a UTSP server. Returns raw, unparsed result file contents.

        :return: electricity, warm water, high bodily activity and low bodily activity result file contents
        :rtype: Tuple[str]
        """
        # Create an LPG configuration and set the simulation parameters
        start_date = self.my_simulation_parameters.start_date.strftime("%Y-%m-%d")
        # Unlike HiSim the LPG includes the specified end day in the simulation --> subtract one day
        last_day = self.my_simulation_parameters.end_date - datetime.timedelta(days=1)
        end_date = last_day.strftime("%Y-%m-%d")
        simulation_config = lpg_helper.create_basic_lpg_config(
            self.utsp_config.household,
            HouseTypes.HT23_No_Infrastructure_at_all,
            start_date,
            end_date,
            self.get_resolution(),
            travel_route_set=self.utsp_config.travel_route_set,
            transportation_device_set=self.utsp_config.transportation_device_set,
            charging_station_set=self.utsp_config.charging_station_set,
            calc_options=[
                CalcOption.HouseholdSumProfilesCsvNoFlex,
                CalcOption.BodilyActivityStatistics,
                CalcOption.TansportationDeviceJsons,
                CalcOption.FlexibilityEvents,
            ],
        )
        assert simulation_config.CalcSpec is not None

        # Enable simulation of transportation and flexible devices
        simulation_config.CalcSpec.EnableTransportation = True
        simulation_config.CalcSpec.EnableFlexibility = True

        # Define required result files
        electricity = result_file_filters.LPGFilters.sum_hh1(
            LoadTypes.Electricity, no_flex=True
        )
        warm_water = result_file_filters.LPGFilters.sum_hh1(
            LoadTypes.Warm_Water, no_flex=True
        )
        high_activity = result_file_filters.LPGFilters.BodilyActivity.HIGH
        low_activity = result_file_filters.LPGFilters.BodilyActivity.LOW
        flexibility = result_file_filters.LPGFilters.FLEXIBILITY_EVENTS
        required_files = {
            f: datastructures.ResultFileRequirement.REQUIRED
            for f in [
                electricity,
                warm_water,
                high_activity,
                low_activity,
                flexibility,
            ]
        }
        # Define transportation result files
        car_states = result_file_filters.LPGFilters.all_car_states_optional()
        car_locations = result_file_filters.LPGFilters.all_car_locations_optional()
        driving_distances = (
            result_file_filters.LPGFilters.all_driving_distances_optional()
        )
        result_files: Dict[str, Optional[datastructures.ResultFileRequirement]] = {
            **required_files,
            **car_states,
            **car_locations,
            **driving_distances,
        }

        # Prepare the time series request
        request = datastructures.TimeSeriesRequest(
            simulation_config.to_json(), "LPG", required_result_files=result_files  # type: ignore
        )

        log.information("Requesting LPG profiles from the UTSP.")
        # Request the time series
        result = client.request_time_series_and_wait_for_delivery(
            self.utsp_config.url, request, self.utsp_config.api_key
        )

        electricity_file = result.data[electricity].decode()
        warm_water_file = result.data[warm_water].decode()
        high_activity_file = result.data[high_activity].decode()
        low_activity_file = result.data[low_activity].decode()
        flexibility_file = result.data[flexibility].decode()

        # Save flexibility and transportation files
        self.save_result_file(flexibility, flexibility_file)
        for filename in itertools.chain(
            car_states.keys(), car_locations.keys(), driving_distances.keys()
        ):
            if filename in result.data:
                self.save_result_file(filename, result.data[filename].decode())

        return electricity_file, warm_water_file, high_activity_file, low_activity_file

    def save_result_file(self, name: str, content: str) -> None:
        """Saves a result file in the folder specified in the config object.

        :param name: the name for the file
        :type name: str
        :param content: the content that will be written into the file
        :type content: str
        """
        filepath = os.path.join(self.utsp_config.result_path, name)
        directory = os.path.dirname(filepath)
        # Create the directory if it does not exist
        try:
            os.makedirs(directory)
        except OSError as exc:
            if exc.errno == errno.EEXIST and os.path.isdir(directory):
                pass
            else:
                raise
        # Create the result file
        with open(filepath, "w", encoding="utf-8") as result_file:
            result_file.write(content)

    def build(self):
        """Retrieves and preprocesses all data for this component."""
        file_exists, cache_filepath = utils.get_cache_file(
            component_key=self.component_name,
            parameter_class=self.utsp_config,
            my_simulation_parameters=self.my_simulation_parameters,
        )
        if file_exists:
            dataframe = pd.read_csv(
                cache_filepath, sep=",", decimal=".", encoding="cp1252"
            )
            self.number_of_residents = dataframe["number_of_residents"].tolist()
            self.heating_by_residents = dataframe["heating_by_residents"].tolist()
            self.electricity_consumption = dataframe["electricity_consumption"].tolist()
            self.water_consumption = dataframe["water_consumption"].tolist()
        else:
            (
                electricity,
                warm_water,
                high_activity,
                low_activity,
            ) = self.get_profiles_from_utsp()

            ################################
            # Calculates heating generated by residents and loads number of residents
            # Heat power generated per resident in W
            # mode 1: awake
            # mode 2: sleeping
            gain_per_person = [150, 100]

            # load occupancy profile
            occupancy_profile = []
            bodily_activity_files = [high_activity, low_activity]
            for filecontent in bodily_activity_files:
                json_filex = json.loads(filecontent)

                occupancy_profile.append(json_filex)

            # see how long csv files from LPG are to check if averaging has to be done and calculate desired length
            steps_original = len(occupancy_profile[0]["Values"])
            simulation_time_span = (
                self.my_simulation_parameters.end_date
                - self.my_simulation_parameters.start_date
            )
            steps_desired = int(
                simulation_time_span.days
                * 24
                * (3600 / self.my_simulation_parameters.seconds_per_timestep)
            )
            steps_ratio = int(steps_original / steps_desired)

            # initialize number of residence and heating by residents
            self.heating_by_residents = [0] * steps_desired
            self.number_of_residents = [0] * steps_desired

            # load electricity consumption and water consumption
            electricity_data = io.StringIO(electricity)
            pre_electricity_consumption = pd.read_csv(
                electricity_data,
                sep=";",
                decimal=".",
                encoding="cp1252",
            )
            water_data = io.StringIO(warm_water)
            pre_water_consumption = pd.read_csv(
                water_data,
                sep=";",
                decimal=".",
                encoding="cp1252",
            )

            # convert electricity consumption and water consumption to desired format and unit
            self.electricity_consumption = pd.to_numeric(
                pre_electricity_consumption["Sum [kWh]"] * 1000 * 60
            ).tolist()  # 1 kWh/min == 60W / min
            self.water_consumption = pd.to_numeric(
                pre_water_consumption["Sum [L]"]
            ).tolist()

            # process data when time resolution of inputs matches timeresolution of simulation
            if steps_original == steps_desired:
                for mode, gain in enumerate(gain_per_person):
                    for timestep in range(steps_original):
                        self.number_of_residents[timestep] += occupancy_profile[mode][
                            "Values"
                        ][timestep]
                        self.heating_by_residents[timestep] = (
                            self.heating_by_residents[timestep]
                            + gain * occupancy_profile[mode]["Values"][timestep]
                        )

            # average data, when time resolution of inputs is coarser than time resolution of simulation
            elif steps_original > steps_desired:
                for mode, gain in enumerate(gain_per_person):
                    for timestep in range(steps_desired):
                        number_of_residents_av = (
                            sum(
                                occupancy_profile[mode]["Values"][
                                    timestep
                                    * steps_ratio : (timestep + 1)
                                    * steps_ratio
                                ]
                            )
                            / steps_ratio
                        )
                        self.number_of_residents[timestep] += np.round(
                            number_of_residents_av
                        )
                        self.heating_by_residents[timestep] = (
                            self.heating_by_residents[timestep]
                            + gain * number_of_residents_av
                        )
                # power needs averaging, not sum
                self.electricity_consumption = [
                    sum(self.electricity_consumption[n : n + steps_ratio]) / steps_ratio
                    for n in range(0, steps_original, steps_ratio)
                ]
                self.water_consumption = [
                    sum(self.water_consumption[n : n + steps_ratio])
                    for n in range(0, steps_original, steps_ratio)
                ]

            else:
                raise Exception(
                    "input from LPG is given in wrong time resolution - or at least cannot be interpolated correctly"
                )

            # Saves data in cache
            data = np.transpose(
                [
                    self.number_of_residents,
                    self.heating_by_residents,
                    self.electricity_consumption,
                    self.water_consumption,
                ]
            )
            database = pd.DataFrame(
                data,
                columns=[
                    "number_of_residents",
                    "heating_by_residents",
                    "electricity_consumption",
                    "water_consumption",
                ],
            )

            database.to_csv(cache_filepath)
            del data
            del database
        self.max_hot_water_demand = max(self.water_consumption)

    def write_to_report(self):
        """Adds a report entry for this component."""
        lines = []
        lines.append(f"Name: {self.component_name}")
        return lines
